# This file is part of pymadcad,  distributed under license LGPL v3

''' Group of functions and math classes of pymadcad '''

import glm
from glm import *
del version, license
import math
from math import pi, inf, nan, atan2
from copy import deepcopy
max = __builtins__['max']
min = __builtins__['min']
any = __builtins__['any']
all = __builtins__['all']
round = __builtins__['round']

from arrex import typedlist
import arrex.glm

# alias definitions
vec2 = dvec2
mat2 = dmat2
vec3 = dvec3
mat3 = dmat3
vec4 = dvec4
mat4 = dmat4
quat = dquat

# numerical precision of floats used
NUMPREC = 1e-13	# float64 here, so 14 decimals
#NUMPREC = 1e-6	# float32 here, so 7 decimals, so 1e-6 when exponent is 1
COMPREC = 1-NUMPREC



# common base definition, for end user
O = vec3(0,0,0)
X = vec3(1,0,0)
Y = vec3(0,1,0)
Z = vec3(0,0,1)


def isfinite(x):
	''' Return false if x contains a `inf` or a `nan` '''
	if isinstance(x, (int,float)):
		return math.isfinite(x)
	return not (glm.any(isinf(x)) or glm.any(isnan(x)))

def norminf(x):
	''' Norm L infinite  ie.  `max(abs(x), abs(y), abs(z))` '''
	return max(glm.abs(x))
# norm L1  ie.  `abs(x) + abs(y) + abs(z)`
norm1 = l1Norm
# norm L2  ie.  `sqrt(x**2 + y**2 + z**2)`   the usual distance also known as euclidian distance or manhattan distance
norm2 = length

def anglebt(x,y) -> float:
	''' Angle between two vectors 
	
		The result is not sensitive to the lengths of x and y
	'''
	n = length(x)*length(y)
	return acos(min(1,max(-1, dot(x,y)/n)))	if n else 0
	
def arclength(p1, p2, n1, n2):
	''' Length of an arc between p1 and p2, normal to the given vectors in respective points '''
	c = max(0, dot(n1,n2))
	if abs(c-1) < NUMPREC:	return 0
	v = p1-p2
	return sqrt(dot(v,v) / (2-2*c)) * acos(c)

def project(vec, dir) -> vec3:
	''' Component of `vec` along `dir`, equivalent to :code:`dot(vec,dir) / dot(dir,dir) * dir` 
	
		The result is not sensitive to the length of `dir`
	'''
	try:	return dot(vec,dir) / dot(dir,dir) * dir
	except ZeroDivisionError:	
		if dot(vec,vec):		return vec3(nan)
		else:					return vec3(0)
		
	
def noproject(vec, dir) -> vec3:
	''' Components of `vec` not along `dir`, equivalent to :code:`vec - project(vec,dir)` 
	
		The result is not sensitive to the length of `dir`
	'''
	return vec - project(vec,dir)

def unproject(vec, dir) -> vec3:
	''' Return the vector in the given direction as if `vec` was its projection on it, equivalent to :code:`dot(vec,vec) / dot(vec,dir) * dir` 
	
		The result is not sensitive to the length of `dir`
	'''
	try:	return dot(vec,vec) / dot(vec,dir) * dir
	except ZeroDivisionError:	
		if dot(vec,vec):		return vec3(nan)
		else:					return vec3(0)

def perpdot(a:vec2, b:vec2) -> float:
	''' Dot product of a with perpendicular vector to b, equivalent to `dot(a, prep(b))` '''
	return -a[1]*b[0] + a[0]*b[1]

def perp(v:vec2) -> vec2:
	''' Perpendicular vector to the given vector '''
	return vec2(-v[1], v[0])
	
def dirbase(dir, align=vec3(1,0,0)):
	''' Return a base using the given direction as z axis (and the nearer vector to align as x) '''
	x = noproject(align, dir)
	if not length2(x) > NUMPREC**2:
		align = vec3(align[2],-align[0],align[1])
		x = noproject(align, dir)
	if not length2(x) > NUMPREC**2:
		align = vec3(align[1],-align[2],align[0])
		x = noproject(align, dir)
	x = normalize(x)
	y = cross(dir, x)
	return x,y,dir

def scaledir(dir, factor=None) -> mat3:
	''' Return a mat3 scaling in the given direction, with the given factor (1 means original scale) 
		If factor is None, the length of dir is used, but it can leads to precision loss on direction when too small.
	'''
	if factor is None:
		factor = length(dir)
		dir = dir / factor
	return mat3(1) + (factor-1)*mat3(dir[0]*dir, dir[1]*dir, dir[2]*dir)
	
def rotatearound(angle, *args) -> mat4:
	''' Return a transformation matrix for a rotation around an axis
		
		rotatearound(angle, axis)
		rotatearound(angle, origin, dir)
	'''
	if len(args) == 1:		origin, dir = args[0]
	elif len(args) == 2:	origin, dir = args
	else:
		raise TypeError('invalid use of rotatearound')
	
	r = mat3_cast(angleAxis(angle, dir))
	m = mat4(r)
	m[3] = vec4(origin - r*origin, 1)
	return m

def transform(*args) -> mat4:
	''' Create an affine transformation matrix.
		
		Supported inputs:
			:mat4:                                    obviously return it unmodified
			:float:                                   scale using the given ratio 
			:vec3:                                    translation only
			:quat, mat3, mat4:                        rotation only
			:(vec3,vec3), (vec3,mat3), (vec3,quat):   `(o,T)` translation and rotation
			:(vec3,vec3,vec3):                        `(x,y,z)` base of vectors for rotation
			:(vec3,vec3,vec3,vec3):                   `(o,x,y,z)` translation and base of vectors for rotation
	'''
	if len(args) == 1 and isinstance(args[0], (tuple,list)):
		args = args[0]
	if len(args) == 1:
		if isinstance(args[0], mat4):	return args[0]
		elif isinstance(args[0], (mat3, quat, int, float)):	return mat4(args[0])
		elif isinstance(args[0], vec3):	return translate(args[0])
	elif len(args) == 2:
		if isinstance(args[0], vec3):
			if   isinstance(args[1], (mat3, quat, int, float)):		m = mat4(args[1])
			elif isinstance(args[1], vec3):		m = mat4(quat(args[1]))
			m[3] = vec4(args[0], 1)
			return m
	elif isinstance(args[0], vec3) and len(args) == 3:			
		return mat4(mat3(*args))
	elif isinstance(args[0], vec3) and len(args) == 4:
		m = mat4(mat3(*args[1:]))
		m[3] = vec4(args[0], 1)
		return m
	
	raise TypeError('a transformation must be a  mat3, mat4, quat, (O,mat3), (O,quat), (0,x,y,z), not {}'.format(args))
	
def transformer(trans):
	''' Return an function to apply the given transform on vectors
		
		Supported inputs:
			:float:	scale by the given ratio
			:vec3:  translate the given position
			:mat3:  rotate the given position
			:quat:  rotate the given position
			:mat4:  affine transform (rotate then translate)
	'''
	if isinstance(trans, (dquat, fquat)):		trans = mat3_cast(trans)
	if callable(trans):													return trans
	if isinstance(trans, (dvec3, fvec3)):								return lambda v: v + trans
	if isinstance(trans, (dmat3, fmat3, dmat4, fmat4, int, float)):		return lambda v: trans * v
	raise TypeError('a transformer must be a  vec3, quat, mat3, mat4 or callable, not {}'.format(trans))



def interpol1(a, b, x):
	''' 1st order polynomial interpolation '''
	return (1-x)*a + x*b

def interpol2(a, b, x):
	''' 3rd order polynomial interpolation 
		a and b are iterable of successive derivatives of a[0] and b[0]
	'''
	return (	2*x*(1-x)  * interpol1(a[0],b[0],x)		# linear component
			+	x**2       * (b[0] + (1-x)*b[1])		# tangent
			+	(1-x)**2   * (a[0] + x*a[1])	# tangent
			)

hermite = spline = interpol2

def intri_flat(pts, a,b):
	A,B,C = pts
	c = 1-a-b
	return a*A + b*B + c*C

def intri_sphere(pts, ptangents, a,b, etangents=None):
	''' Cubic interpolation over a triangle (2 dimension space), edges are guaranteed to fit an interpol2 curve using the edge tangents
	
	.. note::
		If the tangents lengths are set to the edge lengths, that version gives a result close to a sphere surface
	'''
	A,B,C = pts
	ta,tb,tc = ptangents
	c = 1-a-b
	P = a*A + b*B + c*C +  a*b*c * (ta[0] + ta[1] + tb[0] + tb[1] + tc[0] + tc[1])
	return (	0
			+	a**2 * (A + b*ta[0] + c*ta[1])
			+	b**2 * (B + c*tb[0] + a*tb[1])
			+	c**2 * (C + a*tc[0] + b*tc[1])
			+	2*(b*c + c*a + a*b) * P
			)

def intri_smooth(pts, ptangents, a,b):
	''' Cubic interpolation over a triangle, edges are guaranteed to fit an interpol2 curve using the edge tangents
	
	.. note::
		If the tangents lengths are set to the edge lengths, that version gives a result that only blends between the curved edges, a less bulky result than `intri_sphere`
	'''
	A,B,C = pts
	ta,tb,tc = ptangents
	c = 1-a-b
	return (	0
			+	a**2 * (A + b*ta[0] + c*ta[1] + b*c*(ta[0]+ta[1]))
			+	b**2 * (B + c*tb[0] + a*tb[1] + c*a*(tb[0]+tb[1]))
			+	c**2 * (C + a*tc[0] + b*tc[1] + a*b*(tc[0]+tc[1]))
			+	2*(b*c + c*a + a*b) * (a*A + b*B + c*C)
			)

def intri_parabolic(pts, ptangents, a,b, etangents=None):
	''' Quadratic interpolation over a triangle, edges are NOT fitting an interpol2 curve '''
	A,B,C = pts
	ta,tb,tc = ptangents
	c = 1-a-b
	return (	0
			+	a * (A + b*ta[0] + c*ta[1])
			+	b * (B + c*tb[0] + a*tb[1])
			+	c * (C + a*tc[0] + b*tc[1])
			)


# distances:

distance_pp = distance

def distance_pa(pt, axis):
	''' Point - axis distance '''
	return length(noproject(pt-axis[0], axis[1]))

def distance_pe(pt, edge):
	''' Point - edge distance '''
	dir = edge[1]-edge[0]
	l = length2(dir)
	if not l:	return 0
	x = dot(pt-edge[0], dir)/l
	if   x < 0:	return distance(pt,edge[0])
	elif x > 1:	return distance(pt,edge[1])
	else:
		return length(noproject(pt-edge[0], dir))

def distance_aa(a1, a2):
	''' Axis - axis distance '''
	return length(project(a1[0]-a2[0], cross(a1[1], a2[1])))

def distance_ae(axis, edge):
	''' Axis - edge distance '''
	x = axis[1]
	z = normalize(cross(x, edge[1]-edge[0]))
	y = cross(z,x)
	s1 = dot(edge[0]-axis[0], y)
	s2 = dot(edge[1]-axis[0], y)
	if s1*s2 < 0:
		return dot(edge[0]-axis[1], z)
	elif abs(s1) < abs(s2):
		return distance_pa(edge[0], axis)
	else:
		return distance_pa(edge[1], axis)
		
def distance_pt(p, triangle):
	''' Point - triangle distance '''
	normal = cross(triangle[1]-triangle[0], triangle[2]-triangle[0])
	for i in range(3):
		if dot(p-triangle[i-1], cross(triangle[i-2]-triangle[i-1], normal)) > 0:
			return distance_pe(p, (triangle[i-1],triangle[i-2]))
	return length(project(normal, p-triangle[0]))


#-- algorithmic functions ---------



def fbisect(f, start, stop, prec=None):
	''' bisection over the parameter of a continuous real function, returning the place where the function switches from True to False
		f(x) -> bool
	'''
	if not prec:	prec = abs(stop-start)*1e-3
	
	if not f(start):	return start
	elif f(stop):		return stop

	while abs(stop-start) > prec:
		x = (start+stop)*0.5
		if f(x):	start = x
		else:		stop = x
	return start

def bisect(array, value, key=None):
	if key is None:		key = lambda x:x

	start, stop = 0, len(array)
	while start < stop:
		i = (start+stop)//2
		v = key(array[i])
		if v > value:	stop = i
		elif v < value:	start = i+1
		else:	return i
	return start

# TODO rename it first
def find(iterator, predicate, default=None):
	for e in iterator:
		if predicate(e):	return e
	return default
			
def imax(iterable, default=None):
	''' Return the index of the max of the iterable '''
	best = default
	score = -inf
	for i,o in enumerate(iterable):
		if o >= score:
			score = o
			best = i
	if best is None:	raise IndexError('iterable is empty')
	return best

def linstep(start, stop, x):
	''' like smoothstep but with a linear ramp between `start` and `stop` '''
	if x <= start:	return 0
	if x >= stop:	return 1
	return (x-start)/(stop-start)

def linrange(start, stop=None, step=None, div=0, end=True):
	''' Yield successive intermediate values between start and stop 
		
		stepping:
		
		- if `step` is given, it will be the amount between raised value until it gets over `stop`
		- if `div` is given, it will be the number of intermediate steps between `start` and `stop` (with linear spacing)
		
		ending:
		
		- if `end` is True, it will stop iterate with value `stop` (or just before)
		- if `end` is False, it will stop iterating just before `stop` and never with `stop`
		
		Example:
		
			>>> list(linrange(5, -5, div=1))
			[5, 0, -5]
			
			>>> list(linrange(5, -5, div=10)
			
		
		NOTE:  
			If step is given and is not a multiple of `stop-start` then `end` has no influence
	'''
	if stop is None:	start, stop = 0, start
	if step is None:	step = (stop-start)/(div+1)
	elif step * (stop-start) < 0:	step = -step
	if not end:			stop -= step
	stop += NUMPREC*stop
	
	t = start
	while (stop-t)*step >= 0:
		yield t
		t += step


# aliases, for those who like them
Vector = Point = vec3


class Axis(object):
	''' A 3D (zeroed) axis with an origin and a direction
	
		Mathematically speaking, a 3D axis doesn't necessarily have an origin, since any point on it can be its start, but for implementation and convenience reasons this axis has
		
		.. note::
		
			in previous madcad versions, axis were often tuples and not instances of this class. This is why this class has a `__getitem__` allowing to be used like a tuple. But this class should be used instead now.
	'''
	__slots__ = ('origin', 'direction', 'interval')
	def __init__(self, origin, direction=None, interval=None):
		if direction is None:
			origin, direction = vec3(0), origin
		self.origin, self.direction = origin, direction
		self.interval = interval
	
	def __getitem__(self, i):
		''' behave like the axis was a tuple (origin, direction) '''
		if i==0:	return self.origin
		elif i==1:	return self.direction
		else:		raise IndexError('an axis has only 2 components')
		
	def flip(self) -> 'Axis':
		''' switch the axis direction '''
		return Axis(self.origin, -self.direction, self.interval)
	
	def offset(self, increment) -> 'Axis':
		''' move the axis origin along its direction '''
		return Axis(self.origin + self.direction*increment, self.direction, self.interval)
		
	def transform(self, transform) -> 'Axis':
		''' move the axis by the given transformation '''
		if isinstance(transform, (float,int)):		return Axis(transform*self.origin, self.direction, self.interval)
		elif isinstance(transform, vec3):			return Axis(transform+self.origin, self.direction, self.interval)
		elif isinstance(transform, (mat3, quat)):	return Axis(transform*self.origin, normalize(transform*self.direction), self.interval)
		elif isinstance(transform, (mat4)):			return Axis(transform*self.origin, normalize(mat3(transform)*self.direction), self.interval)
		raise TypeError('transform must be one of float, vec3, mat3, quat, mat4')
	
	slvvars = ('origin', 'direction')
	def slv_tangent(self, pt):
		return self.direction
		
	def __repr__(self):
		return 'Axis({}, {})'.format(self.origin, self.direction)
	
	def display(self, scene):
		from .displays import AxisDisplay
		return AxisDisplay(scene, (self.origin, self.direction), self.interval)
			
def isaxis(obj):
	''' Return True if the given object is considered to be an axis.
		An axis can be an instance of `Axis` or a tuple `(vec3, vec3)`
	'''
	return isinstance(obj, Axis) or isinstance(obj, tuple) and len(obj)==2 and isinstance(obj[0],vec3) and isinstance(obj[1],vec3)


class Box:
	''' This class describes a box always orthogonal to the base axis, used as convex for area delimitations 
	
		This class is independent from the dimension or number precision of the used vectors. You can for instance have a `Box` of `vec2` as well as a box of `vec3`. However boxes with different vector types cannot interperate.
	
		Attributes:
			
			min:	vector of minimum coordinates of the box (usually bottom left corner)
			max:	vector of maximum coordinates of the box (usually top right corner)
	
	'''
	__slots__ = ('min', 'max')
	def __init__(self, min=None, max=None, center=vec3(0), width=vec3(-inf)):
		if min and max:			self.min, self.max = min, max
		else:					self.min, self.max = center-width/2, center+width/2
	
	@property
	def center(self) -> vec3:
		''' Mid coordinates of the box '''
		return (self.min + self.max) /2
	@property
	def width(self) -> vec3:
		''' Diagonal vector of the box '''
		return self.max - self.min
	
	def corners(self) -> '[vec3]':
		''' Create a list of the corners of the box '''
		c = self.min, self.max
		t = type(self.min)
		return [
			t(c[i&1][0], c[(i>>1)&1][1], c[(i>>2)&1][2])
			for i in range(8)
			]
			
	def volume(self) -> float:
		''' Volume inside '''
		v = 1
		for edge in self.width:
			v *= edge
		return v
		
	def contain(self, point):
		''' Return True if the given point is inside or on the surface of the box '''
		return all(self.min <= point) and all(point <= self.max)
		
	def inside(self, point):
		''' Return True if the given point is strictly inside the box '''
		return all(self.min < point) and all(point < self.max)
	
	def isvalid(self):
		''' Return True if the box defines a valid space (min coordinates <= max coordinates) '''
		return any(self.min <= self.max)
	def isempty(self):
		''' Return True if the box contains a non null volume '''
		return not any(self.min < self.max)
	
	def __add__(self, other):
		if isinstance(other, vec3):		return Box(self.min + other, self.max + other)
		elif isinstance(other, Box):	return Box(other.min, other.max)
		else:
			return NotImplemented
			
	def __iadd__(self, other):
		if isinstance(other, vec3):		
			self.min += other
			self.max += other
		elif isinstance(other, Box):	
			self.min += other.min
			self.max += other.max
		else:
			return NotImplemented
			
	def __sub__(self, other):
		if isinstance(other, vec3):		return Box(self.min - other, self.max - other)
		elif isinstance(other, Box):	return Box(other.min, other.max)
		else:
			return NotImplemented
	
	def __or__(self, other):	return deepcopy(self).union(other)
	def __and__(self, other):	return deepcopy(self).intersection(other)
	
	def union_update(self, other) -> 'self':
		''' Extend the volume of the current box to bound the given point or box '''
		if isinstance(other, (dvec3, fvec3)):
			self.min = glm.min(self.min, other)
			self.max = glm.max(self.max, other)
		elif isinstance(other, Box):
			self.min = glm.min(self.min, other.min)
			self.max = glm.max(self.max, other.max)
		else:
			raise TypeError('unable to integrate {}'.format(type(other)))
		return self
	
	def intersection_update(self, other) -> 'self':
		''' Reduce the volume of the current box to the intersection between the 2 boxes '''
		if isinstance(other, Box):
			self.min = glm.max(self.min, other.min)
			self.max = glm.min(self.max, other.max)
		else:
			raise TypeError('expected a Box'.format(type(other)))
		return self
	
	def union(self, other) -> 'Box':
		''' Return a box containing the current and the given box (or point) 
		
			Example:
				
				>>> Box(vec2(1,2), vec2(2,3)) .union(vec3(1,4))
				Box(vec2(1,2), vec2(2,4))
				
				>>> Box(vec2(1,2), vec2(2,3)) .union(Box(vec3(1,-4), vec3(2,8)))
				Box(vec2(1,-4), vec3(2,8))
		
		'''
		if isinstance(other, (dvec3, fvec3)):
			return Box(	glm.min(self.min, other),
						glm.max(self.max, other))
		elif isinstance(other, Box):
			return Box(	glm.min(self.min, other.min),
						glm.max(self.max, other.max))
		else:
			raise TypeError('unable to integrate {}'.format(type(other)))
	
	def intersection(self, other) -> 'Box':
		''' Return a box for the volume common to the current and the given box 
		
			Example:
			
				>>> Box(vec2(-1,2), vec2(2,3)) .intersection(Box(vec3(1,-4), vec3(2,8)))
				Box(vec2(1,2), vec3(2,3))
		
		'''
		if isinstance(other, Box):
			return Box(	glm.max(self.min, other.min),
						glm.min(self.max, other.max))
		else:
			raise TypeError('expected a Box'.format(type(other)))
	
	def transform(self, trans) -> 'Box':
		''' Box bounding the current one in a transformed space '''
		if not self.isvalid():	return self
		trans = transformer(trans)
		return boundingbox((trans(p)  for p in self.corners()))
	
	def cast(self, vec) -> 'Box':
		return Box(vec(self.min), vec(self.max))
	
	def __bool__(self):
		return self.isvalid()
	
	def __repr__(self):
		return '{}({}, {})'.format(self.__class__.__name__, repr(self.min), repr(self.max))

def boundingbox(obj, ignore=False, default=Box(width=vec3(-inf))) -> Box:
	''' Return a box containing the object passed
		`obj` can be a vec3, Box, object with a `box()` method, or an iterable of such objects
	'''
	if isinstance(obj, Box):			return obj
	if isinstance(obj, (dvec3, fvec3)):	return Box(obj,obj)
	if hasattr(obj, 'box'):				return obj.box()
	if hasattr(obj, '__iter__'):
		obj = iter(obj)
		if ignore:
			bound = default
			for e in obj:
				try:	bound = boundingbox(e)
				except TypeError:	continue
				break
			for e in obj:
				try:	bound.union(e)
				except TypeError:	continue
		else:
			bound = boundingbox(next(obj, default))
			for e in obj:	bound.union_update(e)
		return bound
	if ignore:
		return default
	else:
		raise TypeError('unable to get a boundingbox from {}'.format(type(obj)))



class Screw(object):
	''' A 3D torsor aka Screw aka Wrench aka Twist - is a mathematical object defined as follow:
		  * a resulting vector R
		  * a momentum vector field M

		The momentum is a function of space, satisfying the relationship:
			M(A) = M(B) + cross(R, A-B)
		
		Therefore it is possible to represent a localized torsor such as:
		  * R = resulting
		  * M = momentum vector at position P
		  * P = position at which M takes the current value
		
		Torsor are useful for generalized solid mechanics to handle multiple variables of the same nature:
		  * Force torsor:	
			  Screw(force, torque, pos)
		  * Velocity (aka kinematic) torsor:
			  Screw(rotation, velocity, pos)
		  * Kinetic (inertia) torsor:
			  Screw(linear movement quantity, rotational movement quantity, pos)
			
		  All these torsors makes it possible to represent all these values independently from expression location
		  
		  
		Attributes:
			resulting (vec3): 
			momentum (vec3):
			position (vec3):
	'''
	__slots__ = ('resulting', 'momentum', 'position')
	def __init__(self, resulting=None, momentum=None, position=None):
		self.resulting, self.momentum, self.position = resulting or vec3(0), momentum or vec3(0), position or vec3(0)
	def locate(self, pt) -> 'Screw':
		''' Gets the same torsor, but expressed for an other location '''
		return Screw(self.resulting, self.momentum + cross(self.resulting, pt-self.position), pt)
	
	def transform(self, mat) -> 'Screw':
		''' Changes the torsor from coordinate system '''
		if isinstance(mat, mat4):
			rot, trans = mat3(mat), vec3(mat[3])
		elif isinstance(mat, mat3):
			rot, trans = mat, 0
		elif isinstance(mat, quat):
			rot, trans = mat, 0
		elif isinstance(mat, vec3):
			rot, trans = 1, mat
		else:
			raise TypeError('Screw.transform() expect mat4, mat3 or vec3')
		return Screw(rot*self.resulting, rot*self.momentum, rot*self.position + trans)
	
	def __add__(self, other):
		if other.position != self.position:		other = other.locate(self.position)
		return Screw(self.resulting+other.resulting, self.momentum+other.momentum, self.position)
	
	def __sub__(self, other):
		if other.position != self.position:		other = other.locate(self.position)
		return Screw(self.resulting-other.resulting, self.momentum-other.momentum, self.position)
	
	def __neg__(self):
		return Screw(-self.resulting, -self.momentum, self.position)
	
	def __mul__(self, x):
		return Screw(x*self.resulting, x*self.momentum, self.position)
	
	def __div__(self, x):
		return Screw(self.resulting/x, self.momentum/x, self.position)
		
	def __repr__(self):
		return '{}(\n\t{}, \n\t{}, \n\t{})'.format(self.__class__.__name__, repr(self.resulting), repr(self.momentum), repr(self.position))

def comomentum(t1, t2):
	''' Comomentum of screws:   `dot(M1, R2)  +  dot(M2, R1)`
		
		The result is independent of torsors location
	'''
	t2 = t2.locate(t1.position)
	return dot(t1.momentum, t2.resulting) + dot(t2.momentum, t1.resulting)





from ..mathutils import typedlist, vec3
from ..hashing import suites, connpe
from .container import NMesh
from .mesh import Mesh
from .web import Web
from .wire import Wire


def mesh(*arg, resolution=None) -> Mesh:
	''' Build a Mesh object from supported objects:
	
		:mesh:              return it with no copy
		:primitive:         call its ``.mesh`` method and convert the result to web
		:iterable:          convert each element to web and join them
	'''
	if not arg:	
		raise TypeError('mesh takes at least one argument')
	if len(arg) == 1:	
		arg = arg[0]
	if isinstance(arg, Mesh):		
		return arg
	elif hasattr(arg, 'mesh'):
		return mesh(arg.mesh(resolution=resolution))
	elif hasattr(arg, '__iter__'):
		pool = Mesh()
		for primitive in arg:
			pool += mesh(primitive, resolution=resolution)
		pool.mergeclose()
		return pool
	else:
		raise TypeError('incompatible data type for Web creation')

def web(*arg, resolution=None) -> Web:
	''' Build a Web object from supported objects:
	
		:web:               return it with no copy
		:wire:              reference points and generate edge couples
		:primitive:         call its ``.mesh`` method and convert the result to web
		:iterable:          convert each element to web and join them
		:list of vec3:      reference it and generate trivial indices
		:iterable of vec3:  get points and generate trivial indices
	'''
	if not arg:	
		raise TypeError('web takes at least one argument')
	if len(arg) == 1:	
		arg = arg[0]
	if isinstance(arg, Web):		
		return arg
	elif isinstance(arg, Wire):	
		return Web(
				arg.points, 
				arg.edges(), 
				arg.tracks[:-1] if arg.tracks else None, 
				groups=arg.groups,
				)
	elif hasattr(arg, 'mesh'):
		return web(arg.mesh(resolution=resolution))
	elif isinstance(arg, (typedlist,list,tuple)) and isinstance(arg[0], vec3):
		return Web(arg, [(i,i+1) for i in range(len(arg)-1)])
	elif hasattr(arg, '__iter__'):
		pool = Web()
		for primitive in arg:
			pool += web(primitive, resolution=resolution)
		pool.mergeclose()
		return pool
	else:
		raise TypeError('incompatible data type for Web creation')

def wire(*arg, resolution=None) -> Wire:
	''' Build a Wire object from the other compatible types.
		Supported types are:
		
		:wire:              return it with no copy
		:web:               find the edges to joint, keep the same point buffer
		:primitive:         call its ``.mesh`` method and convert the result to wire
		:iterable:          convert each element to Wire and joint them
		:list of vec3:      reference it and put trivial indices
		:iterable of vec3:  create internal point list from it, and put trivial indices
	'''
	if not arg:	
		raise TypeError('wire takes at least one argument')
	if len(arg) == 1:	
		arg = arg[0]
	if isinstance(arg, Wire):		
		return arg
	elif isinstance(arg, Web):
		indices = suites(arg.edges)
		if len(indices) > 1:	raise ValueError('the given web has junctions or is discontinuous')
		return Wire(arg.points, indices[0], groups=[None])	# TODO: find a way to get the groups from the Web edges through suites or not
	elif hasattr(arg, 'mesh'):
		return wire(arg.mesh(resolution=resolution))
	elif isinstance(arg, vec3):
		return Wire([arg])
	elif isinstance(arg, (typedlist,list,tuple)):
		try:
			return Wire(arg, groups=[None])
		except TypeError:
			pass
	if hasattr(arg, '__iter__'):
		pool = Wire()
		for primitive in arg:
			pool += wire(primitive, resolution=resolution)
		pool.mergeclose()
		return pool
	else:
		raise TypeError('incompatible data type for Wire creation')




# This file is part of pymadcad,  distributed under license LGPL v3

from .mathutils import *
from .kinematic.solver import Joint
from .kinematic.displays import scale_solid, world_solid, kinematic_color
from .mesh import Mesh, Wire, wire, web, Web
from .scheme import Scheme
from . import generation as gt
from . import generation, primitives, settings

__all__ = ['Revolute', 'Planar', 'Prismatic', 'Cylindrical', 'Ball', 'PointSlider', 'EdgeSlider', 'Ring',
			'Cam', 'Contact',
			'Rack', 'Gear', 'Helicoid']


cornersize = 0.1	

def drotate(angle, axis):
	# derivative of sin and cos an argument translation of pi/2
	m = rotate(angle+0.5*pi, axis)
	# remove homogeneous one and axis proper space
	m -= mat4(outerProduct(axis, axis) / length2(axis))
	return m
	
def dtranslate(direction):
	return translate(direction) - mat4()


class Revolute(Joint):
	''' Joint for revolution around an axis
	'''
	bounds = (-inf, inf)
	default = 0
	
	def __init__(self, solids, axis: Axis, local=None, default=None):
		self.solids = solids
		local = local or axis
		
		if default is not None:
			self.default = default
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
			self.position = axis[3].xyz, local[3].xyz
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
			self.position = axis[0], local[0]
		else:
			raise TypeError('Revolute can only be placed on an Axis or a mat4')
		
	def __repr__(self):
		return '{}({}, Axis(({:.3g},{:.3g},{:.3g}), ({:.3g},{:.3g},{:.3g})))'.format(
			self.__class__.__name__, self.solids, *self.post[3].xyz, *self.post[2].xyz)
	
	def direct(self, angle) -> mat4:
		return self.post * rotate(angle, Z) * self.pre
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * mat3(matrix) * affineInverse(self.pre)
		angle = atan2(m[1][0] - m[0][1], m[0][0] + m[1][1])
		angle += (2*pi) * ((angle + pi - close) // (2*pi))
		return angle
		
	def grad(self, angle, delta=1e-6):
		return self.post * drotate(angle, Z) * self.pre
		
	def transmit(self, force, parameters=None, velocity=None) -> Screw:
		l = force.locate(self.axis[0])
		return Screw(l.resulting, project(l.momentum, self.axis[1]))
		
	def scheme(self, index, maxsize, attach_start, attach_end):
		size = settings.display['joint_size']
		radius = size/4
		sch = Scheme()
		resolution = ('div', 16)
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size))
		cylinder = gt.cylinder(
						-z*size*0.4, 
						+z*size*0.4, 
						radius=size/4, 
						resolution=resolution,
						fill=False,
						)
		sch.add(cylinder, shader='ghost')
		sch.add(cylinder.outlines(), shader='line')
		if attach_start:
			v = normalize(noproject(attach_start - o, z))
			if not isfinite(v):
				v = x
			p = v*size/4
			sch.add(gt.cone(p+size*cornersize*v, p, size*cornersize, fill=False, resolution=('div', 4)), shader='fill')
			sch.set(shader='line')
			sch.add(p + v*size*cornersize)
			sch.add(attach_start, space=world_solid(self.solids[0]))
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size))
		c1 = primitives.Circle(
						(-z*size*0.5, -z), 
						radius, 
						resolution=resolution,
						).mesh()
		c2 = primitives.Circle(
						(+z*size*0.5, z), 
						radius, 
						resolution=resolution,
						).mesh()
		# sch.add([z*size*0.5, o+z*size*0.5], shader='line')
		sch.add(c1, shader='line')
		sch.add(c2, shader='line')
		sch.add(gt.flatsurface(c1), shader='ghost')
		sch.add(gt.flatsurface(c2), shader='ghost')
		
		if attach_end:
			sch.set(shader='line')
			side = z * size * 0.5
			if dot(attach_end-o, side) < 0:
				side = -side
			if dot(attach_end, side)**2 < length2(attach_end)*length2(side) * 0.25:
				radial = normalize(noproject(attach_end-o, z))
				sch.add(side + radial*radius)
				sch.add(side + radial*2*radius)
			else:
				sch.add(side)
			sch.add(attach_end, space=world_solid(self.solids[1]))
		
		return sch

		
class Planar(Joint):
	''' Joint for translation in 2 directions and rotation around the third direction 
		
		Classical definition:	Planar (direction vector)
		the initial state requires an additional distance between the solids
		this class holds an axis for each side, the axis origins are constrained to share the same projections on the normal
	'''
		
	bounds = (vec3(-inf), vec3(inf))
	default = vec3(0, 0, 0)
	
	def __init__(self, solids, axis, local=None, default=None):
		self.solids = solids
		local = local or axis
		
		if default is not None:
			self.default = default
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
			self.position = axis[3].xyz, local[3].xyz
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
			self.position = axis[0], local[0]
		else:
			raise TypeError('Planar can only be placed on an Axis or a mat4')
	
	def __repr__(self):
		return '{}({}, Axis(({:.3g},{:.3g},{:.3g}), ({:.3g},{:.3g},{:.3g})))'.format(
			self.__class__.__name__, self.solids, *self.post[3].xyz, *self.post[2].xyz)
	
	def direct(self, position: tuple):
		return mat4(self.post) * translate(vec3(position.xy, 0)) * rotate(position.z, Z) * mat4(self.pre)
		
	def inverse(self, matrix, close=None):
		m = mat4(transpose(self.post)) * matrix * mat4(transpose(self.pre))
		angle = atan2(m[0][1], m[0][0])
		angle += (2*pi) * ((angle + pi - close) // (2*pi))
		return vec3(m[3].xy, angle)
		
	def grad(self, position: vec3, delta=1e-6):
		return (
			self.post * translate(X) * rotate(position.z, Z) * self.pre,
			self.post * translate(Y) * rotate(position.z, Z) * self.pre,
			self.post * translate(vec3(position.xy,0)) * drotate(position.z, Z) * self.pre,
			)
	
	def transmitable(self, force, parameters=None, velocity=None):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0])
		normal = normalize(a0[1] + a1[1])
		return Screw(project(action.resulting, normal), noproject(action.momentum, normal), action.position)
	
	
	def scheme(self, index, maxsize, attach_start, attach_end):		
		size = settings.display['joint_size']
		sch = Scheme()
		gap = 0.1
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size),
			)
		square = gt.parallelogram(x*size, y*size, origin=o-gap*size*z, align=0.5)
		sch.add(square.outlines(), shader='line')
		sch.add(square, shader='ghost')
		
		if attach_start:
			sch.add(gt.cone(o-size*(gap+cornersize)*z, o-size*gap*z, size*cornersize, fill=False, resolution=('div', 4)), shader='fill')
			sch.set(shader='line')
			sch.add(o-size*(gap+cornersize)*z)
			sch.add(attach_start, space=world_solid(self.solids[0]))
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		sch.set(
			track = index[self.solids[1]],
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size),
			)
		square = gt.parallelogram(-x*size, y*size, origin=o+gap*size*z, align=0.5)
		sch.add(square.outlines(), shader='line')
		sch.add(square, shader='ghost')
		
		if attach_end:
			sch.add(gt.cone(o+size*(gap+cornersize)*z, o+size*gap*z, size*cornersize, fill=False, resolution=('div', 4)), shader='fill')
			sch.set(shader='line')
			sch.add(o+size*(gap+cornersize)*z)
			sch.add(attach_end, space=world_solid(self.solids[1]))
		
		return sch


class Prismatic(Joint):
	''' Joint for translation only in a direction 
		
		Classical definition:  Prismatic (direction vector)
		the initial state requires more parameters: the relative placements of the solids
		this class holds a base for each of the solids, bases are (0,X,Y)  where X,Y are constrained to keep the same direction across bases, and the bases origins lays on their common Z axis
	'''
	
	bounds = (-inf, inf)
	default = 0
	
	def __init__(self, solids, axis: Axis, local=None, default=None):
		self.solids = solids
		local = local or axis
		
		if default is not None:
			self.default = default
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
			self.position = axis[3].xyz, local[3].xyz
		elif isinstance(axis, vec3):
			self.post = mat4(quat(Z, axis))
			self.pre = mat4(quat(local, Z))
			self.position = axis, local
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
			self.position = axis[0], local[0]
		else:
			raise TypeError('Track can only be placed on a vec3, Axis or mat4')
	
	def __repr__(self):
		return '{}({}, vec3({:.3g},{:.3g},{:.3g}))'.format(
			self.__class__.__name__, self.solids, *self.post[3].xyz)
	
	def direct(self, translation):
		return self.post * translate(translation*Z) * self.pre
		
	def inverse(self, matrix, close=None):
		return (affineInverse(self.post) * matrix * affineInverse(pre))[2][2]
		
	def grad(self, translation, delta=1e-6):
		return self.post * dtranslate(Z) * self.pre
	
	def transmit(self, force, parameters=None, velocity=None):
		z0, z1 = cross(b0[1],b0[2]), cross(b1[1],b1[2])
		normal = normalize(z0 + z1)
		return Screw(noproject(action.resulting, normal), action.momentum, action.position)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		size = settings.display['joint_size']
		sch = Scheme()
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size),
			)
		profile = gt.parallelogram(0.4*size*x, 0.4*size*y, origin=o, align=0.5, fill=False)
		profile.tracks = typedlist(range(len(profile.edges)))
		exterior = gt.extrusion(profile, size*z, alignment=0.5)
		exterior.splitgroups()
		sch.add(exterior, shader='ghost')
		sch.add(exterior.outlines(), shader='line')
		if attach_start:
			v = attach_start - o
			if abs(dot(v,x)) > abs(dot(v,y)):
				v = x if dot(v,x)>0 else -x
			else:
				v = y if dot(v,y)>0 else -y
			p = o + v*0.2*size
			sch.add(gt.cone(p+size*cornersize*v, p, size*cornersize, fill=False, resolution=('div', 4)), shader='fill')
			sch.set(shader='line')
			sch.add(p+size*cornersize*v)
			sch.add(attach_start, space=world_solid(self.solids[0]))
	
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size),
			)
		interior = gt.extrusion(Web([
			o-0.2*size*(x+y),
			o+0.2*size*(x+y),
			o-0.2*size*(x-y),
			o+0.2*size*(x-y),
			],
			[(0,1), (2,3)],
			), size*z, alignment=0.5)
		sch.add(interior.outlines(), shader='line')
		if attach_end:
			v = attach_end - o
			v = z if dot(v, z) > 0 else -z
			p = o + v*size/2
			sch.add(p)
			sch.add(attach_end, space=world_solid(self.solids[1]))
			
		return sch


class Cylindrical(Joint):
	''' Joint with rotation and translation around an axis
	'''
	bounds = (vec2(-inf), vec2(inf))
	default = vec2(0)
	
	def __init__(self, solids, axis: Axis, local=None, default=None):
		self.solids = solids
		local = local or axis
		
		if default is not None:
			self.default = default
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
			self.position = axis[3].xyz, local[3].xyz
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
			self.position = axis[0], local[0]
		else:
			raise TypeError('Pivot can only be placed on an Axis or a mat4')
		
	def __repr__(self):
		return '{}({}, Axis(({:.3g},{:.3g},{:.3g}), ({:.3g},{:.3g},{:.3g})))'.format(
			self.__class__.__name__, self.solids, *self.post[3].xyz, *self.post[2].xyz)
	
	def direct(self, parameters) -> mat4:
		return self.post * rotate(parameters[0], Z) * translate(parameters[1]*Z) * self.pre
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * mat4(matrix) * affineInverse(self.pre)
		angle = atan2(m[1][0] - m[0][1], m[0][0] + m[1][1])
		angle += (2*pi) * ((angle + pi - close) // (2*pi))
		return angle, m[2][2]
		
	def grad(self, parameters, delta=1e-6):
		return (
			self.post * drotate(parameters[0], Z) * self.pre,
			self.post * dtranslate(Z) * self.pre,
			)
		
	def transmit(self, force, parameters=None, velocity=None) -> Screw:
		l = force.locate(self.axis[0])
		return Screw(l.resulting, project(l.momentum, self.axis[1]))
		
	def scheme(self, index, maxsize, attach_start, attach_end):
		size = settings.display['joint_size']
		radius = size/4
		sch = Scheme()
		resolution = ('div', 16)
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size))
		cylinder = gt.cylinder(
						o-z*size*0.4, 
						o+z*size*0.4, 
						radius=size/4, 
						resolution=resolution,
						fill=False,
						)
		sch.add(cylinder, shader='ghost')
		sch.add(cylinder.outlines(), shader='line')
		if attach_start:
			v = normalize(noproject(attach_start - o, z))
			if not isfinite(v):
				v = x
			p = v*size/4
			sch.add(gt.cone(p+size*cornersize*v, p, size*cornersize, fill=False, resolution=('div', 4)), shader='fill')
			sch.set(shader='line')
			sch.add(p + v*size*cornersize)
			sch.add(attach_start, space=world_solid(self.solids[0]))
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size))
		sch.add([o-z*size*0.5, o+z*size*0.5], shader='line')
		
		sch.set(shader='fill')
		if attach_end:
			sch.set(shader='line')
			side = z * size * 0.5
			if dot(attach_end-o, side) < 0:
				side = -side
			if dot(attach_end, side)**2 < length2(attach_end)*length2(side) * 0.25:
				radial = normalize(noproject(attach_end-o, z))
				sch.add(side + radial*radius)
				sch.add(side + radial*2*radius)
			else:
				sch.add(side)
			sch.add(attach_end, space=world_solid(self.solids[1]))
			
		return sch

class Ball(Joint):
	''' Joint for rotation all around a point.
	
		Classical definition: Ball (point)
		the initial state doen't require more data
		the class holds a point for each side
	'''
	def __init__(self, solids, center, local=None, default=None):
		self.solids = solids
		local = local or center
		
		if default is not None:
			self.default = default
		
		if isinstance(center, mat4):
			self.post = center
			self.pre = affineInverse(local)
			self.position = center[3].xyz, local[3].xyz
		elif isinstance(center, vec3):
			self.post = translate(center)
			self.pre = translate(-local)
			self.position = center, local
		else:
			raise TypeError("ball can only be placed on a vec3 or mat4")
	
	bounds = (quat(-2,-2,-2,-2), quat(2,2,2,2))
	default = quat()
			
	def normalize(self, orient):
		return normalize(quat(orient))
	
	def direct(self, orient):
		return self.post * mat4(normalize(quat(orient))) * self.pre
	
	def inverse(self, matrix, close=None):
		return quat(affineInverse(self.post) * matrix * affineInverse(self.pre))
		
	def grad(self, orient):
		a,b,c,d = orient
		return (
			self.post * mat4( 
				 2*a,  2*d, -2*c, 0,
				-2*d,  2*a,  2*b, 0,
				 2*c, -2*b,  2*a, 0,
				 0,   0,     0,   0,
				) * self.pre,
			self.post * mat4(
				 2*b,  2*c,  2*d, 0,
				 2*c, -2*b,  2*a, 0,
				 2*d, -2*a, -2*b, 0,
				 0,    0,    0,   0,
				) * self.pre,
			self.post * mat4(
				-2*c,  2*b, -2*a, 0,
				 2*b,  2*c,  2*d, 0,
				 2*a,  2*d, -2*c, 0,
				 0,    0,    0,   0,
				) * self.pre,
			self.post * mat4(
				-2*d,  2*a,  2*b, 0,
				-2*a, -2*d,  2*c, 0,
				 2*b,  2*c,  2*d, 0,
				 0,    0,    0,   0,
				) * self.pre,
			)
	
	def transmit(self, force, parameters=None, velocity=None):
		return Screw(vec3(0), force.momentum, self.center)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		from .primitives import ArcCentered
		
		size = settings.display['joint_size']
		radius = size/3
		sch = Scheme()
		resolution = ('div', 16)
		
		center = self.post[3].xyz
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(center), maxsize/size))
		if attach_start:
			x,y,z = dirbase(normalize(center - attach_start))
		else:
			x,y,z = mat3()
		profile = ArcCentered(Axis(O,y), x*radius, -z*radius).mesh(resolution=resolution)
		emisphere = gt.revolution(profile, Axis(O,z), resolution=resolution)
		sch.add(emisphere, shader='ghost')
		sch.add(emisphere.outlines(), shader='line')
		if attach_start:
			sch.set(shader='line')
			sch.add(-z*radius)
			sch.add(attach_start, space=world_solid(self.solids[0]))
		
		center = affineInverse(self.pre)[3].xyz
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(center), maxsize/size))
		sch.add(gt.icosphere(O, radius*0.8, resolution=('div', 3)), shader='ghost')
		if attach_end:
			sch.set(shader='line')
			sch.add(radius*0.8*normalize(attach_end-center))
			sch.add(attach_end, space=world_solid(self.solids[1]))
		
		return sch


class PointSlider(Joint):
	''' Joint for rotation all around a point belonging to a plane.
	
		Classical definition: Punctiform/Sphere-Plane (axis)
		The initial state does not require more data
		The class holds a normal axis for the plane side, and a point for the sphere side (that defaults to the axis' origin)
	'''
	def __init__(self, solids, axis, local=None):
		self.solids = solids
		self.axis = axis
		local = local or axis
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
		else:
			raise TypeError("PointSlider can only be placed on Axis or mat4")
	
	default = [0]*5
	bounds = ([-inf]*5, [inf]*5)
	
	def direct(self, parameters):
		translation = vec2(parameters[0:2])
		rotation = quat(parameters[2:5])
		return self.post * translate(vec3(translation, 0)) * mat4(rotation) * self.pre
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * matrix * affineInverse(self.pre)
		return (
			*vec2(m[3]),
			*quat(m),
			)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		size = settings.display['joint_size']
		radius = size * 0.2
		resolution = ('div', 16)
		sch = Scheme()
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size),
			)
		square = gt.parallelogram(x*size, y*size, origin=o+radius*z, align=0.5)
		sch.add(square.outlines(), shader='line')
		sch.add(square, shader='ghost')
		
		if attach_start:
			attach = o+radius*z
			sch.add([attach, attach+z*0.1*size, attach+(x+y)*0.1*size], shader='fill')
			sch.add([attach+cornersize*size*z, attach_start], shader='line')
		
		center = affineInverse(self.pre)[3].xyz
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(center), maxsize/size),
			)
		sch.add(gt.icosphere(center, radius, resolution), shader='ghost')
		
		if attach_end:
			x,y,z = dirbase(attach_end)
			attach = radius*z
			sch.add([attach, attach+z*cornersize*size, attach+(x+y)*cornersize*size], shader='fill')
			sch.add([attach+cornersize*size*z, attach_start], shader='line')
		
		return sch


class EdgeSlider(Joint):
	''' Joint for rotation all around a point belonging to a plane.
	
		Classical definition: Punctiform/Sphere-Plane (axis)
		The initial state does not require more data
		The class holds a normal axis for the plane side, and a point for the sphere side (that defaults to the axis' origin)
	'''
	def __init__(self, solids, axis, local=None):
		self.solids = solids
		self.axis = axis
		local = local or axis
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
		else:
			raise TypeError("EdgeSlider can only be placed on Axis or mat4")
	
	default = [0]*4
	bounds = ([-inf]*4, [inf]*4)
	
	def direct(self, parameters):
		translation = vec2(parameters[0:2])
		rotation = quat(vec3(*parameters[2:4], 0))
		return self.post * translate(vec3(translation, 0)) * mat4(rotation) * self.pre
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * matrix * affineInverse(self.pre)
		q = quat(m)
		return (
			*vec2(m[3]),
			pitch(q),
			roll(q),
			)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		size = settings.display['joint_size']
		radius = 0.2*size
		resolution = ('div', 16)
		sch = Scheme()
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size),
			)
		square = gt.parallelogram(x*size, y*size, origin=o+radius*z, align=0.5)
		sch.add(square.outlines(), shader='line')
		sch.add(square, shader='ghost')
		
		if attach_start:
			attach = o+radius*z
			sch.add([attach, attach+cornersize*size*z, attach+cornersize*size*(x+y)], shader='fill')
			sch.add([o+cornersize*size*z, attach_start], shader='line')
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size),
			)
		cylinder = gt.cylinder(o-x*0.5*size, o+x*0.5*size, radius, fill=False, resolution=resolution)
		sch.add(cylinder, shader='ghost')
		sch.add(cylinder.outlines(), shader='line')
		
		if attach_end:
			x,y,z = dirbase(noproject(attach_end, x))
			attach = radius*z
			sch.add([attach, attach+z*cornersize*size, attach+(x+y)*cornersize*size], shader='fill')
		
		return sch
	

class Ring(Joint):
	''' ring joint '''
	def __init__(self, solids, center, local=None):
		self.solids = solids
		local = local or center
		
		if isinstance(center, mat4):
			self.pre = center
			self.post = local
		elif isinstance(center, vec3):
			self.pre = translate(-center)
			self.post = translate(local)
		else:
			raise TypeError("ball can only be placed on a vec3 or mat4")
	
	bounds = ((-inf, -1, -1, -1, -1), (inf, 1, 1, 1, 1))
	default = (0, 1, 0, 0, 0)
	
	def direct(self, orient):
		return self.post * translate(params[0]) * mat4(quat(params[1:5])) * self.pre
	
	def inverse(self, matrix, close=None):
		return quat(affineInverse(self.post) * matrix * affineInverse(self.pre))
	
	def transmit(self, force, parameters=None, velocity=None):
		return Screw(vec3(0), force.momentum, self.center)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		from .primitives import ArcCentered
		
		size = settings.display['joint_size']
		radius = size/3
		sch = Scheme()
		resolution = ('div', 16)
		
		center = self.post[3].xyz
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(center), maxsize/size))
		if attach_start:
			x,y,z = dirbase(attach_end - center)
		else:
			x,y,z = mat3()
		profile = ArcCentered(Axis(center,y), center+x*radius, center-z*radius).mesh(resolution=resolution)
		emisphere = gt.revolution(profile, Axis(center,z), resolution)
		sch.add(emisphere, shader='ghost')
		sch.add(emisphere.outlines(), shader='line')
		
		center = affineInverse(self.pre)[3].xyz
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(center), maxsize/size))
		sch.add(gt.icosphere(center, radius*0.8, resolution), shader='ghost')
		
		return sch

	
class Universal(Joint):
	''' universal joint '''
	def __init__(self, solids, axis, local=None):
		self.solids = solids
		local = local or axis
		
		if isinstance(axis, mat4):
			self.post = axis
			self.pre = affineInverse(local)
		elif isinstance(axis, Axis):
			self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
			self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
		else:
			raise TypeError('Universal joint can only be placed on an Axis or a mat4')
	
	bounds = (vec2(-inf), vec2(inf))
	default = vec2(0)
	
	def direct(self, orient):
		return self.post * mat4(quat(vec3(orient, 0))) * self.pre
	
	def inverse(self, matrix, close=None):
		m = quat(affineInverse(self.post) * matrix * affineInverse(self.pre))
		return vec2(pitch(m), yaw(m))
		
	def grad(self, orient):
		pitch, yaw = orient
		return (
			self.post * drotate(pitch, X) * self.pre,
			self.post * drotate(pitch, Y) * self.pre,
			)
			
	def scheme(self, index, maxsize, attach_start, attach_end):
		from .primitives import ArcCentered
		
		size = settings.display['joint_size']
		radius = size/3
		sch = Scheme()
		resolution = ('div', 16)
		
		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = scale_solid(self.solids[0], fvec3(o), maxsize/size))
		if attach_start:
			o = self.post[3].xyz
			x,y,z = dirbase(attach_end - center, x)
			sch.add(attach_start, space=world_solid(self.solids[0]))
			sch.add(o - z*radius)
		sch.add(ArcCentered(Axis(o,x), y*radius, -y*radius).mesh(resolution), shader='line')
		sch.add([y, -y])
		
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = scale_solid(self.solids[1], fvec3(o), maxsize/size))
		x,y,z,o = affineInverse(self.pre)
		if attach_end:
			o = affineInverse(self.pre)[3].xyz
			x,y,z = dirbase(attach_end - center, y)
			sch.add(attach_end, space=world_solid(self.solids[1]))
			sch.add(o - z*radius)
		sch.add(ArcCentered(Axis(o,y), x*radius, -x*radius).mesh(resolution), shader='line')
		sch.add([x, -x])
		
		return sch
		
	
	
class ConstantVelocity(Joint):
	''' not implemented yet '''
	pass
	

class Project(Joint):
	''' not implemented yet '''
	def __init__(self, solids, projection:mat4, target:mat4):
		indev

	
class Rack(Joint):
	''' Rack to gear interaction
		
		`radius` is the ratio between the gear rotation and the rack translation
		
		Attributes:
			radius:  ratio between the gear rotation and the rack translation
			rack:    rack axis tangent to the gear
			axis:    gear axis
	'''
	def __init__(self, solids, radius:float, rack:Axis, axis):
		self.solids = solids
		self.radius = radius
		self.axis = axis
		self.rack = rack
		self.pre = mat4(quat(axis[1], Z)) * translate(-axis[0])
		self.post = translate(rack[0]) * mat4(quat(X, rack[1]))
		self.position = self.rack[0], self.axis[0]
		
	bounds = (-inf, inf)
	default = 0
	
	def direct(self, translation):
		return self.post * translate(translation*X) * rotate(-translation/self.radius,Z) * self.pre
	
	def grad(self, translation):
		return (
			+ self.post * dtranslate(X) * rotate(-translation/self.radius,Z) * self.pre
			+ self.post * translate(translation*X) * drotate(-translation/self.radius,Z)/-self.radius * self.pre
			)
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * matrix * affineInverse(self.pre)
		return atan(*m[0].xy)
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		from .generation import revolution
		from .primitives import Circle, Segment
		sch = Scheme()

		x,y,z,o = dmat4x3(self.post)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = world_solid(self.solids[0]))
		sch.add([o-self.radius*y-self.radius*x, o-self.radius*y+self.radius*x], shader='line')
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		h = 0.1*self.radius*z
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = world_solid(self.solids[1]))
		sch.add(Circle(Axis(o,z), self.radius).mesh(), shader='line')
		sch.add(revolution(Segment(o+self.radius*x+h, o+self.radius*x-h), Axis(o,z)), shader='ghost')
		
		return sch

		
	

class Gear(Joint):
	''' Gear interaction between two solids.
	
		`ratio` is the factor from the rotation of s1 to the rotation of s2
		The two pinions are considered circular, but no assumption is made on their radius or relative inclination.
		The interaction is symetric.
		
		Attributes:
			ratio:  the gear rotation transmission ratio
			centerline: the relative positioning of the gears axis, could be 'float', `vec3`, `mat3`, `quat`, `mat4`
			axis:  first gear axis
			local:  second gear axis
	'''
	def __init__(self, solids, ratio:float, centerline:mat4, axis, local):
		self.solids = solids
		self.ratio = ratio
		if isinstance(centerline, (int,float)):
			centerline = centerline*X
		self.centerline = transform(centerline)
		self.axis = axis
		self.pre = mat4(quat(local[1], Z)) * translate(-local[0])
		self.post = translate(axis[0]) * mat4(quat(Z, axis[1]))
		self.position = self.axis[0], local[0]
		
	bounds = (-inf, inf)
	default = 0
		
	def direct(self, angle):
		return self.post * rotate(angle,Z) * self.centerline * rotate(-self.ratio*angle,Z) * self.pre
		
	def grad(self, angle):
		return (
			+ self.post * drotate(angle,Z) * self.centerline * rotate(-self.ratio*angle,Z) * self.pre
			+ self.post * rotate(angle,Z) * self.centerline * drotate(-self.ratio*angle,Z)*-self.ratio * self.pre
			)
		
	def inverse(self, matrix, close=None):
		m = affineInverse(self.post) * matrix * affineInverse(self.pre)
		angle = atan(*m[0].xy) / (1+self.ratio)
		return (angle - close + pi) % (2*pi) - pi
	
	def scheme(self, index, maxsize, attach_start, attach_end):
		from .generation import revolution
		from .primitives import Circle, Segment
		sch = Scheme()
		
		# radial vector
		r0 = length(self.centerline[3].xy) / abs(self.centerline[2].z - 1/self.ratio)
		r1 = r0 / abs(self.ratio)
		
		x,y,z,o = dmat4x3(self.post)
		h = 0.1*min(r0, r1) * (z * r0 + self.centerline[2].xyz * r1) / (r0 + r1)
		sch.set(
			track = index[self.solids[0]], 
			color = kinematic_color(index[self.solids[0]]),
			space = world_solid(self.solids[0]))
		sch.add(Circle(Axis(o,z), r0).mesh(), shader='line')
		sch.add(revolution(Segment(o+r0*x+h, o+r0*x-h), Axis(o,z)), shader='ghost')
		
		x,y,z,o = dmat4x3(affineInverse(self.pre))
		h = 0.1*min(r0, r1) * (self.centerline[2].xyz * r0 + z * r1) / (r0 + r1)
		sch.set(
			track = index[self.solids[1]], 
			color = kinematic_color(index[self.solids[1]]),
			space = world_solid(self.solids[1]))
		sch.add(Circle(Axis(o,z), r1).mesh(), shader='line')
		sch.add(revolution(Segment(o+r1*x+h, o+r1*x-h), Axis(o,z)), shader='ghost')
		
		return sch


class Helicoid(Joint):
	''' Screw a solid into an other.
	
		`step` is the translation distance needed for that one solid turn by 2*pi around the other
		The interaction is symetric.
		
		not implemented yet
	'''
	def __init__(self, s0, s1, step, b0, b1=None, position=None):
		self.step = step	# m/tr
		self.solids = s0, s1
		if isaxis(b0):			b0 = b0[0], *dirbase(b0[1])[:2]
		if b1 and isaxis(b1):	b1 = b1[0], *dirbase(b1[1])[:2]
		self.bases = b0, b1 or b0
		self.position = position or (self.bases[0][0], self.bases[1][0])
	
	def direct(self, depth):
		indev
		
	def inverse(self, matrix, close=None):
		indev
	
	
	def scheme(self, solid, size, junc):
		if solid is self.solids[0]:
			radius = size/4
			o,x,y = self.bases[0]
			z = cross(x,y)
			center = o + project(self.position[0]-o, z)
			cyl = generation.extrusion(
						web(primitives.Circle(
							(center-z*size*0.5, z), 
							radius, 
							resolution=('div', 16),
							)),
						z*size, 
						)
			l = len(cyl.points)
			v = junc - center
			v = normalize(noproject(v,z))
			if glm.any(isnan(v)):
				v,_,_ = dirbase(z)
			p = center + v*size/4
			cyl.points.append(p)
			cyl.points.append(p + z*size*cornersize)
			cyl.points.append(p + v*size*cornersize)
			cyl.points.append(junc)
			
			return Scheme(
					cyl.points, 
					cyl.faces, 
					[(l,l+1,l+2)], 
					(	[(l, l+2), (l+2, l+3)] 
					+	[(i-1,i) for i in range(1,l//2)] 
					+	[(i-1,i) for i in range(l//2+1,l)] 
					+	[(l//2-1,0), (l-1,l//2)]),
					)
					
		elif solid is self.solids[1]:
			radius = size/6
			o,x,y = self.bases[1]
			z = cross(x,y)
			center = o + project(self.position[1]-o, z)
			attach = side = z * size
			
			heli = []
			div = 32
			a = 4*pi
			for i in range(-div,div+1):
				t = i/div
				heli.append( radius*(cos(a*t)*x + sin(a*t)*y) + center + side*t )
			heli.extend([center+side, junc])
			heli = web(Wire(heli))
			return Scheme(heli.points, [], [], heli.edges)


class Cam(Joint):
	''' not implemented yet '''
	pass
	
	
class Contact(Joint):
	''' not implemented yet '''
	pass




# This file is part of pymadcad,  distributed under license LGPL v3

''' This module defines the types and functions for kinematic manimulation and computation.

	
	Kinematics are a conceptual approach of mechanisms. It sort parts in groups called solids (in solids all parts have the same movement), and links the solids to each other using constraints named joints.
	That way no matter what are the parts, or what are their shape, or how parts interact - solids movements can be deduced only from joints.

	This allows designing the mechanisms before designing its parts. This also allows visualizing the mechanism whether it is complete or not.
	
	As parts in the same solid all have the same movement, solids are considered to be undeformable. This allows the to use the Screw theory to represent the force and movement variables (see https://en.wikipedia.org/wiki/Screw_theory). 
	In this module, screws are called ``Screw``.
	
	This module mainly features:
		- `Joint` - the base class for all joints, instances of joints define a kinematic
		- `Kinematic` - the general kinematic solver
		- `Chain` - a joint and kinematic solver dedicated to kinematic chains
			
	joints are defined in `madcad.joints`
'''

from .solver import *
from .assembly import *



# This file is part of pymadcad,  distributed under license LGPL v3
'''
This module focuses on automated envelope generations, based on interface outlines.

The user has only to define the interface outlines or surfaces to join, and the algorithm makes the surface. No more pain to imagine some fancy geometries.
	
Formal definitions
------------------
	
:interface:   a surface or an outline (a loop) with associated exterior normals.
:node:        a group of interfaces meant to be attached together by a blended surface.

In order to generate envelopes, this module asks for cutting all the surfaces to join into 'nodes'. The algorithm decides how to join shortly all the outlines in a node. Once splited in nodes, you only need to generate node junctions for each, and concatenate the resulting meshes.

Details
-------

The blended surfaces are created between interfaces, linked as the points of a convex polyedron of the interfaces directions to the node center.

	
Example
-------

.. code-block:: python

	>>> x, y, z = vec3(1, 0, 0), vec3(0, 1, 0), vec3(0, 0, 1)

	>>> # 1. A surface can be passed: the surface outlines and 
	>>> # normals will be used as the generated surface tangents 
	>>> # 2. Wire or primitives can be passed: the wire loops are used 
	>>> # and the approximate normal to the  wire plane
	>>> # 3. More parameters when building directly the interface
	>>> m = junction(
	...		extrusion(2 * z, web(Circle((vec3(0), z), 1))), # 1.
	...		Circle((2 * x, x), 0.5),                        # 2.
	...		(Circle((2 * y, y), 0.5), "tangent", 2.0),      # 3.
	...		generate="normal",
	... )

To come in a next version

.. code-block:: python
	
	>>> # create junction for each iterable of interface, 
	>>> # if some are not interfaces, they are used 
	>>> # as placeholder objects for auto-determined interfaces
	>>> multijunction(
	...		(surf1, surf2, 42, surf5),
	...		(42, surf3, surf4),
	...		generate='straight',
	... )
'''

from .mesh import Mesh, Wire, Web, wire, mkquad, numpy_to_typedlist, typedlist_to_numpy
from .hashing import connef, edgekey, suites, arrangeface
from .mathutils import *
from . import settings
from . import generation

from .nprint import nprint


def get_interfaces(objs, tangents, weights):
	''' Collects the data definig interfaces for junctions and blendings out of an iterable of tuples 
		This function has no real interest for the enduser.
	'''
	# simply merge results of interface_parse
	v_pts = []
	v_tangents = []
	v_weights = []
	loops = []
	for obj in objs:
		# parse
		args = [None, tangents, weights, None]
		if isinstance(obj, tuple):
			for i,o in enumerate(obj):	args[i] = o
		else:
			args[0] = obj
		e_pts, e_tangents, e_weights, e_loops = get_interface(*args)
		# concatenate
		l = len(v_pts)
		v_pts.extend(e_pts)
		v_tangents.extend(e_tangents)
		v_weights.extend(e_weights)
		loops.extend([i+l  for i in loop] for loop in e_loops)
	return v_pts, v_tangents, v_weights, loops
		
def get_interface(base, tangents, weights, loops):
	''' Collects the data definig interfaces for junctions and blendings out of one object 
		This function has no real interest for the enduser.
	'''
	# get the interface outline to connect to the others
	if loops is None:
		if not isinstance(base, (Wire, Web, Mesh)) and hasattr(base, 'mesh'):	
			base = base.mesh()
		if not isinstance(base, (Wire, Web, Mesh)):
			raise TypeError('expected one of Wire,Web,Mesh   and not {}'.format(type(base).__name__))
		
		base.strippoints()
		points = base.points
		if isinstance(base, Wire):		loops = [base.indices]
		elif isinstance(base, Web):		loops = suites(base.edges)
		elif isinstance(base, Mesh): 
			loops = suites(base.outlines_oriented())
	else:
		if not isinstance(base, list):
			raise TypeError('if loops are provided, points must be a list')
		points = base
	
	# normal to each point
	# tangents provided explicitely
	if isinstance(tangents, (list,dict)):
		tangents = tangents
	# one tangent shared by all points
	elif isinstance(tangents, vec3):
		tangents = [tangents] * len(points)
		
	elif tangents == 'straight':
		tangents = [vec3(0)] * len(points)
	
	elif tangents == 'normal':
		# tangents normal to surface
		if isinstance(base,Mesh):
			tangents = base.vertexnormals()
		# tangents are in the common direction of adjacent faces
		else:
			tangents = [None] * len(points)
			for loop in loops:
				for i,n in zip(loop, Wire(points, loop).vertexnormals(True)):
					tangents[i] = n
	
	elif tangents == 'tangent':
		tangents = [None] * len(points)
		# tangents tangents to surface outline
		if isinstance(base,Mesh):
			for p,t in base.tangents().items():
				tangents[p] = t
		# tangents are tangents to a guessed surface in the loop
		else:
			for loop in loops:
				for i,n in zip(loop, Wire(points, loop).tangents(True)):
					tangents[i] = n
	else:
		raise ValueError('wrong tangents specification')
	
	# factor applied on each tangents
	if not isinstance(weights, list):
		weights = [weights] * len(points)
	return points, tangents, weights, loops



def junction(*args, center=None, tangents='normal', weight=1., match='length', resolution=None):
	''' Join several outlines with a blended surface
		
		tangents:	
			'straight'	no interpolation, straight lines
			'normal'	interpolated surface starts normal to the interfaces
			'tangent'	interpolated surface starts tangent to the interfaces
		
		weight:
			factor applied on the tangents before passing to `interpol2` or `intri_smooth`
			the resulting tangent is computed in point `a` as `weight * distance(a,b) * normalize(tangent[a])`
			
		match:
			'length'	share the outline between 2 matched points to assign the same length to both sides
			'corner'	split the curve to share around a corner
			
		center:		
			position of the center of the junction node used to determine connexion between interfaces
			can be usefull for particularly weird and ambiguous interfaces
			
		.. note::
			match method 'corner' is not yet implemented
	'''
	pts, tangents, weights, loops = get_interfaces(args, tangents, weight)
	if len(loops) == 0:
		return Mesh()
	if len(loops) == 1:	
		return blendloop(
					(pts, tangents, weights, loops), 
					center, resolution=resolution)
	if len(loops) == 2:
		c0 = interfaces_center(pts, loops[0])
		c1 = interfaces_center(pts, loops[1])
		z = c1 - c0
		p = pts[imax( length2(noproject(pts[i]-c0, z))   for i in loops[0] )]
		x = noproject(p - c0, z)
		i0 = imax( dot(pts[i], x)  for i in loops[0])
		i1 = imax( dot(pts[i], x)  for i in loops[1])
		return blendpair(
					(pts, tangents, weights, (
						loops[0][i0:]+loops[0][1:i0+1],
						loops[1][i1:]+loops[1][1:i1+1],
						)),
					resolution=resolution)
	
	# determine center and convex hull of centers
	if not center:
		center = interfaces_center(pts, *loops)
	node = convexhull([	normalize(interfaces_center(pts, interf)-center) 
						for interf in loops])
	# fix convexhull faces orientations
	for i,f in enumerate(node.faces):
		if dot(node.facenormal(f), sum(node.facepoints(i))) < -1e-9:
			node.faces[i] = (f[0],f[2],f[1])
	nodeconn = connef(node.faces)
	
	# determine the junction triangles and cuts
	middles = []
	cuts = {}
	for face in node.faces:
		n = node.facenormal(face)
		middle = [	max(loops[i], key=lambda p: dot(pts[p], n) )	
					for i in face]
		#middle = [None]*3
		#for i in range(3):
			#interface = loops[face[i]]
			#middle[i] = interface[max(range(len(interface)), 
				#key=lambda j: dot(pts[interface[i]], n) * dot(n, 
									#cross(	pts[interface[j]] - pts[interface[j-1]], 
											#node.points[face[i]] - node.points[face[i-1]])),
				#)]
		middles.append(middle)
		for i in range(3):
			if middle[i-1] in cuts and cuts[middle[i-1]] != middle[i-2]:	continue
			cuts[middle[i-1]] = middle[i]
	# cut the interfaces
	parts = {}
	for interf in loops:
		l = len(interf)
		last = None
		first = None
		for i,p in enumerate(interf):
			if p in cuts:
				if first is None:	first = i
				if last is not None:
					parts[interf[last]] = interf[last:i+1]
				last = i
		if (first-last) % len(interf) > 1:
			parts[interf[last]] = interf[last:] + interf[1:first+1]
	
	# assemble parts in matchings
	matchs = []
	done = set()
	for i in cuts:
		if i in done:	continue
		j = parts[cuts[i]][-1]
		assert parts[cuts[j]][-1] == i, "failed to match outlines"
		matchs.append((parts[cuts[i]], parts[cuts[j]]))
		done.add(j)
	
	# generate the surfaces
	result = Mesh(pts)
	div = 11
	for la,lb in matchs:
		def infos():
			for a,b in match_length(Wire(pts,lb), Wire(pts,list(reversed(la)))):
				# normalize except for 0
				ta, tb = tangents[a], tangents[b]
				ta /= length(ta) or 1
				tb /= length(tb) or 1
				# scale and weight
				l = distance(pts[a],pts[b])  # NOTE not the arclength for the moment
				yield (pts[a], ta*l*weights[a]), (pts[b], tb*l*weights[b])
		result += blenditer(infos(), div, interpol2)
	for tri in middles:
		assert len(tri) == 3
		ptgts = [None]*3
		ptri = [None]*3
		for i in range(3):
			s = tri[i]
			ptgts[i] = (tangents[s]*weights[s]*distance(pts[s], pts[tri[i-2]]), 
						tangents[s]*weights[s]*distance(pts[s], pts[tri[i-1]]),
						)
			ptri[i] = pts[tri[i]]
			
		result += generation.dividedtriangle(lambda u,v: intri_smooth(ptri, ptgts, u,v), div)
	
	result.mergeclose()
	return result
	


def multijunction(*nodes, **kwargs):
	''' Resolves partial definitions of interfaces, then create junctions with it 
	
		nodes are lists of interfaces, each of these list will end in a call to `junction`
		nodes can contain 
		- interfaces (or objects than can be casted into)
		- couples of the following format: (key, mix)
			key - hashable objects used to implicitely define an interface
			mix - float used to set the interface center
	'''
	indev

	
def interfaces_center(pts, *loops):
	center = vec3(0)
	total = 0
	for loop in loops:
		for i in range(len(loop)):
			l = length(pts[loop[i-1]]+pts[loop[i]])
			center += (pts[loop[i-1]] + pts[loop[i]]) * l/2
			total += l
	return center / total if total else center

def convexhull(pts):
	import scipy.spatial
	if len(pts) == 3:
		return Mesh(pts, [(0,1,2),(0,2,1)])
	else:
		return Mesh(pts, numpy_to_typedlist(
			scipy.spatial.ConvexHull(
				typedlist_to_numpy(pts, 'f8'), 
				qhull_options='QJ Pp',
				).simplices, uvec3))
	


def match_length(line1, line2) -> '[(int, int)]':
	''' Yield couples of point indices where the curved absciss are the closest '''
	yield line1.indices[0], line2.indices[0]
	l1, l2 = line1.length(), line2.length()
	i1, i2 = 1, 1
	x1, x2 = 0, 0
	while i1 < len(line1.indices) and i2 < len(line2.indices):
		p1 = distance(line1.points[line1.indices[i1-1]], line1.points[line1.indices[i1]]) / l1
		p2 = distance(line2.points[line2.indices[i2-1]], line2.points[line2.indices[i2]]) / l2
		x1p = x1+0.5*p1
		x2p = x2+0.5*p2
		if x1 <= x2p and x2p <= x1+p1    and    x2 <= x1p and x1p <= x2+p2:
			i1 += 1; x1 += p1
			i2 += 1; x2 += p2
		elif x1p < x2p:	
			i1 += 1; x1 += p1
		else:				
			i2 += 1; x2 += p2
		yield line1.indices[i1-1], line2.indices[i2-1]
	while i1 < len(line1.indices):
		i1 += 1
		yield line1.indices[i1-1], line2.indices[i2-1]
	while i2 < len(line2.indices):
		i2 += 1
		yield line1.indices[i1-1], line2.indices[i2-1]


def match_closest(line1, line2) -> '[(int, int)]':
	''' Yield couples of points by cutting each line at the curvilign absciss of the points of the other '''
	l1, l2 = line1.length(), line2.length()
	p1, p2 = line1.points[line1.indices[0]], line2.points[line2.indices[0]]
	x = 0
	i1, i2 = 1, 1
	yield (p1, p2)
	while i1 < len(line1.indices) and i2 < len(line2.indices):
		n1 = line1.points[line1.indices[i1]]
		n2 = line2.points[line2.indices[i2]]
		dx1 = distance(p1, n1) / l1
		dx2 = distance(p2, n2) / l2
		if dx1 > dx2:
			x += dx2
			p1 = interpol1(p1, n1, dx2/dx1)
			p2 = n2
			i2 += 1
		else:
			x += dx1
			p1 = n1
			p2 = interpol1(p2, n2, dx1/dx2)
			i1 += 1
		yield (p1, p2)

def dividematch_length(line1, line2, resolution=None, associated=None) -> '[(vec3, vec3)]':
	''' Insert additional couples to ensure smoothness '''
	indev

def dividematch_closest(w0:Wire, w1:Wire, resolution=None, associated=None) -> (Wire, Wire):
	indev

def blend(interfaces, generate='straight', match='length', resolution=None) -> 'Mesh':
	''' Create a direct blended surface between unclosed lines '''	
	pts, tangents, weights, loops = get_interfaces(args, tangents, weight)
	
	indev
	
def blendloop(interface, center=None, tangents='tangent', weight=1., resolution=None) -> Mesh:
	''' Blend inside a loop interface 
	 
		see `junction` for the parameters.
	'''
	pts, tangents, weights, loops = get_interfaces([interface], tangents, weight)
	if len(loops) > 1:	
		raise ValueError('interface must have one loop only')
	loop = loops[0]
	
	# get center and normal at center
	if not center:
		center = interfaces_center(pts, *loops) + sum(weights[p]*tangents[p] for p in loop) / len(loop)
	normal = normalize(sum(normalize(center-pts[p])	for p in loop))
	if not isfinite(normal):	normal = vec3(0)
	
	# generatestraight
	match = [None] * len(loop)
	div = 0
	for i,p in enumerate(loop):
		match[i] = m = (	(pts[p], distance(center, pts[p])*weights[p]*tangents[p]), 
							(center, noproject(pts[p]-center, normal))	)
		div = max(div, settings.curve_resolution(distance(m[0][0],m[1][0]), anglebt(m[0][1],m[1][1]), resolution))
	return blenditer(match, div, interpol2)


def blendpair(*interfaces, match='length', tangents='tangent', weight=1., resolution=None) -> Mesh:
	''' Blend between a pair of interfaces 
		
		match:   
			`'length'`, `'closest'` refer to `match_*` in this module 
	 
		see `junction` for the other parameters.
	'''
	pts, tangents, weights, loops = get_interfaces(interfaces, tangents, weight)
	if len(loops) != 2:	
		raise ValueError('interface must have exactly 2 loops, got '+str(len(loops)))
	
	if match == 'length':		method = match_length
	elif match == 'closest':	method = match_closest
	else:
		raise ValueError('matching method {} not implemented'.format(match))
	
	#if loops[1][0] == loops[1][-1]:		loops[1] = synchronize(Wire(pts, loops[0]), Wire(pts, loops[1])).indices
	#elif loops[0][0] == loops[0][-1]:	loops[0] = synchronize(Wire(pts, loops[1]), Wire(pts, loops[0])).indices
	
	match = list(method(Wire(pts, loops[0]), Wire(pts, list(reversed(loops[1])))))
	# get the discretisation
	div = 0
	for i in range(1, len(match)):
		a,d = match[i-1][0], match[i-1][1]
		b,c = match[i][0], match[i][1]
		m = (pts[a] + pts[b] + pts[c] + pts[d]) /4
		ta = pts[a] -m
		tb = pts[b] -m
		tc = pts[c] -m
		td = pts[d] -m
		div = max(
				div,
				settings.curve_resolution(
						distance(pts[a],pts[c]), 
						anglebt(ta, -tc) + anglebt(ta, tangents[a]) + anglebt(tc, tangents[c]), 
						resolution),
				settings.curve_resolution(
						distance(pts[b],pts[d]), 
						anglebt(tb, -td) + anglebt(tb, tangents[b]) + anglebt(td, tangents[d]), 
						resolution),
				)
	def infos():
		for p0, p1 in match:
			d = distance(pts[p0], pts[p1])
			yield ((pts[p0], d*weights[p0]*tangents[p0]), (pts[p1], d*weights[p1]*tangents[p1]))
	return blenditer(infos(), div, interpol2)

def blenditer(parameters, div, interpol) -> Mesh:
	''' Create a blended surface using the matching parameters and the given interpolation 
		parameters is an iterable of tuples of arguments for the interpolation function
		interpol receive the elements iterated and the interpolation position at the end
	'''
	segts = div+2
	mesh = Mesh(groups=[None])
	# create interpolation points
	steps = 0
	for params in parameters:
		steps += 1
		for i in range(segts):
			x = i/(segts-1)
			mesh.points.append(interpol(*params, x))
	# create faces
	for i in range(steps-1):
		j = i*segts
		for k in range(segts-1):
			s = j+k
			mkquad(mesh, (s, s+1, s+segts+1, s+segts))
	mesh.mergeclose()
	return mesh


def synchronize(ref:Wire, loop:Wire) -> Wire:
	''' Assuming loop is a closed Wire, it will change its start point to match the startpoint of ref 
	'''
	r0, r1 = 0, int(wire_atlength(ref, ref.length()/2))	# lookup points on ref
	l0 = l1 = 0	# lookup points on loop: start and middle
	s0 = s1 = 0	# offset distance of lookup points on loop
	half = wire_atlength(loop, loop.length()/2)
	l1 = int(half)
	s1 = distance(loop[l1-1], loop[l1]) * (half - l1)
	
	# take the l0,l1 couple that minimize the squared distance between matching lookup points
	best = distance2(ref[r0], loop[l0]) + distance2(ref[r1], loop[l1])
	candidate = l0
	for l0 in range(1,len(loop)):
		# increment l0
		incr = distance(loop[l0-1], loop[l0])
		s0 += incr
		# increment l1
		while True:
			incr = distance(loop[l1], loop[(l1+1)%len(loop)])
			if abs(s1-s0+incr) > abs(s1-s0):
				break
			s1 += incr
			l1 = (l1+1) % len(loop)
		# search for a minimum
		score = distance2(loop[l0], ref[r0]) + distance2(loop[l1], ref[r1])
		if score < best:
			candidate = l0
			best = score
	l0 = candidate
	
	# circulate the loop
	end = loop.indices[l0:]
	if loop[0] == loop[-1]:		end.pop()
	return Wire(loop.points, end + loop.indices[:l0+1])

def wire_atlength(wire, length):
	''' Return the index of the wire point at the given length.
	
		the returned value is float whose integer part is the index and floating part is the 
		ratio of the next segment to use to reach the exact desired length.
	''' 
	s = 0
	for i in range(1,len(wire)):
		d = distance(wire[i-1], wire[i])
		s += d
		if s > length:
			return i - (length-s)/d
	
def join(mesh, line1, line2):
	''' Simple straight surface created from matching couples of line1 and line2 using mesh indices for lines '''
	group = len(mesh.groups)
	mesh.groups.append(None)
	match = iter(curvematch(Wire(mesh.points, line1), Wire(mesh.points, line2)))
	last = next(match)
	for couple in match:
		a,b = last
		d,c = couple
		if b == c:		mktri(mesh, (a,b,c), group)
		elif a == d:	mktri(mesh, (a,b,c), group)
		else:
			mkquad(mesh, (a,b,c,d), group)
		last = couple
		
def trijoin(pts, ptgts, div):
	''' Simple straight surface created between 3 points, interpolation is only on the sides '''
	mesh = Mesh(groups=[None])
	segts = div+1
	for i in range(3):
		for j in range(segts):
			x = j/segts
			mesh.points.append(interpol2((pts[i-1], ptgts[i-1][0]), (pts[i], ptgts[i][1]), x))
	l = len(mesh.points)
	for i in range(3):
		c = i*segts
		for j in range(segts//2):
			mkquad(mesh, ((c-j-1)%l, (c-j)%l, (c+j)%l, (c+j+1)%l))
	return mesh



# This file is part of pymadcad,  distributed under license LGPL v3

''' This module provides annotation functions to quickly measure and show things on meshes
	
	This is more to be considered as a rendering feature, rather than a proper measuring system.
	Everything is built on the class `Scheme` that provide a very versatile way to structure and render simple but animated schematics
	
	Those functions are designed to be very simple to use, (sometimes even minimalistic)
		
		>>> mesh = brick(width=vec3(3,2,1))
		>>> show([
		... 	mesh,
		... 	note_distance_planes(mesh.group(4), mesh.group(2)),
		... 	note_leading(mesh.group(3), text='truc'),
		... 	])
			
'''

import moderngl as mgl
import glm
from operator import itemgetter
from collections import deque
from dataclasses import dataclass
import numpy.linalg
import numpy as np

from .mathutils import *
from .rendering import Display
from .common import resourcedir
from .mesh import Mesh, Web, Wire, web, wire, mesh_distance
from .hashing import connef, connpe, connpp, edgekey, arrangeface, arrangeedge
from .rendering import Displayable, writeproperty, overrides
from .text import TextDisplay, textsize
from .primitives import *
from . import mathutils
from . import generation as gt
from . import settings

__all__ = ['Scheme', 
			'note_leading', 
			'note_floating', 
			'note_label', 
			'note_distance', 
			'note_angle', 
			'note_distance_planes',
			'note_distance_set',
			'note_angle_planes', 
			'note_angle_edge',
			'note_radius',
			'note_bounds',
			]


class Scheme:
	''' An object containing schematics. 
	
		This is a buffer object, it isnot intended to be useful to modify a scheme.
		
		Attributes:
		
			spaces (list):       a space is any function giving a mat4 to position a point on the screen (openGL convensions as used)
			
			vertices (list):	    
			
				a vertex is a tuple
				
				`(space id, position, normal, color, layer, track, flags)`
			
			primitives (list):   
			
				list of buffers (of point indices, edges, triangles, depending on the exact primitive type), associaded to each supported shader in the scheem
				
				currently supported shaders are:
				
				- `'line'`  uniform opaque/transparent lines
				- `'fill'`  uniform opaque/transparent triangles
				- `'ghost'` triangles of surface that fade when its normal is close to the view
			
			components (list):   objects to display setting their local space to one of the spaces
			
			current (dict):     last vertex definition, implicitely reused for convenience
	'''
	def __init__(self, vertices=None, spaces=None, primitives=None, **kwargs):
		self.vertices = vertices or [] # list of vertices
		self.spaces = spaces or []	# definition of each space
		self.primitives = primitives or {} # list of indices for each shader
		self.components = []	# displayables associated to spaces
		# for creation: last vertex inserted
		self.current = {'color':fvec4(settings.colors['annotation'],1), 'flags':0, 'layer':0, 'space':world, 'shader':'line', 'track':0, 'normal':fvec3(0)}
		self.continuous = False
		self.set(**kwargs)
		
		# TODO: regroup spaces when duplicates exists
		
	def __iadd__(self, other):
		''' Concatenante the content of an other scheme in the current one
			
			the current settings stay as before concatenation
		'''
		ls = len(self.spaces)
		lv = len(self.vertices)
		self.spaces.extend(other.spaces)
		self.components.extend(other.components)
		self.vertices.extend([
							v[0] + ls, 
							*v[1:5],
							v[5], 
							v[6],
						]  for v in other.vertices)
		for shader, prims in other.primitives.items():
			if shader not in self.primitives:
				self.primitives[shader] = []
			self.primitives[shader].extend(tuple(i+lv  for i in p) for p in prims)
		
		return self
		
	def __add__(self, other):
		''' Return the union of the two schemes.
		
			The current settings are those from first scheme
		'''
		r = deepcopy(self)
		r += other
		return r
		
	def set(self, *args, **kwargs):
		''' Change the specified attributes in the current default vertex definition '''
		if args:
			if len(args) == 1 and isinstance(args[0], dict):
				kwargs = args[0]
			else:
				raise TypeError('Scheme.set expects keywords argument or one unique dictionnary argument')
		self.current.update(kwargs)
		# register the space if not already known
		if not isinstance(self.current['space'], int):
			try:	i = self.spaces.index(self.current['space'])
			except ValueError:	
				i = len(self.spaces)
				self.spaces.append(self.current['space'])
			self.current['space'] = i
		if not isinstance(self.current['color'], fvec4):
			self.current['color'] = fvec4(self.current['color'])
		
		return self
	
	def add(self, obj, **kwargs):
		''' Add an object to the scheme.
			If it is a mesh, it's merged in the current buffers. 
			Else it is added as a component to the current space.
		'''
		self.set(kwargs)
		if self.current['shader'] not in self.primitives:
			self.primitives[self.current['shader']] = indices = []
		else:
			indices = self.primitives[self.current['shader']]
		l = len(self.vertices)
		continuing = False
		
		if isinstance(obj, (Mesh,Web)):
			self.vertices.extend([
								self.current['space'], 
								fvec3(p), 
								self.current['normal'], 
								self.current['color'], 
								self.current['layer'], 
								self.current['track'], 
								self.current['flags'],
							]  for p in obj.points)
		if isinstance(obj, Mesh):
			indices.extend(((a+l, b+l, c+l)  for a,b,c in obj.faces))
			for i,n in enumerate(obj.vertexnormals()):
				self.vertices[i+l][2] = n
		elif isinstance(obj, Web):
			indices.extend(((a+l, b+l)  for a,b in obj.edges))
		elif isinstance(obj, vec3):
			self.vertices.append([
								self.current['space'], 
								fvec3(obj), 
								self.current['normal'], 
								self.current['color'], 
								self.current['layer'], 
								self.current['track'], 
								self.current['flags'],
								])
			continuing = True
			if self.continuous:
				indices.append((l-1, l))
		
		elif hasattr(obj, '__iter__'):
			n = len(self.vertices)
			for obj in obj:
				if isinstance(obj, (fvec3, vec3)):
					self.vertices.append([
								self.current['space'], 
								fvec3(obj), 
								self.current['normal'], 
								self.current['color'], 
								self.current['layer'], 
								self.current['track'], 
								self.current['flags'],
								])
					n += 1
				else:
					self.add(obj)
			indices.extend((i,i+1)	for i in range(l, n-1))
		else:
			self.component(obj)
			
		self.continuous = continuing
		return self
	
	def component(self, obj, **kwargs):
		''' Add an object as component associated to the current space '''
		self.set(**kwargs)
		self.components.append((self.current['space'], obj))
		
		return self
	
	class display(Display):
		''' Display for schemes
			
			Attributes:
			:spaces:       numpy array of matrices for each space, sent as uniform to the shader
			:vb_vertices:  vertex buffer for vertices
			:vas:          vertex array associated to each shader
		'''
		max_spaces = 4096  # this maximum size of the spaces array must match the size in the shader
		
		@dataclass
		class Shader:
			program: mgl.Program
			mode: int
			enable: int
			blend_func: int
		
		def __init__(self, scene, sch):
			ctx = scene.ctx
			
			# load the resources
			self.shaders, self.shader_ident = scene.resource('scheme', self.load)
			
			# switch to array indexed spaces
			self.spaces = typedlist.full(fmat4(0), len(sch.spaces))
			self.spacegens = list(sch.spaces)
			if len(self.spacegens) > self.max_spaces:
				raise OverflowError('the number of local spaces exceeds the arbitrary build-in limit of {}'.format(self.max_spaces))
			
			self.components = [(space,scene.display(obj))	for space,obj in sch.components]
			self._selected = False
			
			self.nidents = max(v[5] for v in sch.vertices)+1
			self.box = boundingbox(
						(fvec3(v[1]) for v in sch.vertices if self.spacegens[v[0]] is world), 
						default=Box(center=fvec3(0), width=fvec3(-inf)),
						)
			
			# prepare the buffer of vertices
			vertices = np.empty(len(sch.vertices), 'u1, 3f4, 3f4, 4u1, f4, u2, u1')
			for i,v in enumerate(sch.vertices):
				vertices[i] = (
					*v[:3],
					u8vec4(v[3]*255), 
					*v[4:]
					)			
			self.vb_vertices = ctx.buffer(vertices)
			self.b_spaces = ctx.buffer(self.spaces, dynamic=True)
			verticesdef = [(self.vb_vertices, 'u1 3f4 3f4 4f1 f4 u2 u1', 
								'space', 
								'v_position', 
								'v_normal', 
								'v_color', 
								'v_layer', 
								'v_ident', 
								'v_flags')]
			
			# prepare the rending commands
			ident_triangles = []
			ident_lines = []
			self.vas = {}
			self.vai_triangles = None
			self.vai_lines = None
			for shname,batch in sch.primitives.items():
				if not batch:	continue
				if shname not in self.shaders:	raise KeyError('no shader for name {}'.format(repr(shname)))
				
				shader = self.shaders[shname]
				vb_indices = ctx.buffer(np.array(batch, 'u4'))
				self.vas[shname] = ctx.vertex_array(shader.program, verticesdef, vb_indices, mode=shader.mode, skip_errors=True)
				if shader.mode == mgl.LINES:			ident_lines.extend(batch)
				elif shader.mode == mgl.TRIANGLES:		ident_triangles.extend(batch)
			
			if ident_triangles:	self.vai_triangles	= ctx.vertex_array(self.shader_ident, verticesdef, ctx.buffer(np.array(ident_triangles, 'u4')), skip_errors=True)
			if ident_lines:		self.vai_lines 		= ctx.vertex_array(self.shader_ident, verticesdef, ctx.buffer(np.array(ident_lines, 'u4')), skip_errors=True)
			
		def __del__(self):
			for va in self.vas.values():
				va.release()
			self.vas.clear()
			if self.vai_triangles:
				self.vai_triangles.release()
			if self.vai_lines:
				self.vai_lines.release()
			if self.vb_vertices:
				self.vb_vertices.release()
				self.vb_vertices = None
			
		def load(self, scene):
			''' Load shaders and all static data for the current OpenGL context '''
			vert = open(resourcedir+'/shaders/scheme.vert').read()
			shader_ident = scene.ctx.program(
						vertex_shader=vert,
						fragment_shader=open(resourcedir+'/shaders/scheme-ident.frag').read(),
						)
			shaders = {
				'line': self.Shader(
					mode = mgl.LINES, 
					enable = mgl.BLEND | mgl.DEPTH_TEST,
					blend_func = (mgl.SRC_ALPHA, mgl.ONE_MINUS_SRC_ALPHA),
					program = scene.ctx.program(
						vertex_shader=vert,
						fragment_shader=open(resourcedir+'/shaders/scheme-uniform.frag').read(),
						)),
				'fill': self.Shader(
					mode = mgl.TRIANGLES,
					enable = mgl.BLEND | mgl.DEPTH_TEST | mgl.CULL_FACE,
					blend_func = (mgl.SRC_ALPHA, mgl.ONE_MINUS_SRC_ALPHA),
					program = scene.ctx.program(
						vertex_shader=vert,
						fragment_shader=open(resourcedir+'/shaders/scheme-uniform.frag').read(),
						)),
				'ghost': self.Shader(
					mode = mgl.TRIANGLES,
					enable = mgl.BLEND | mgl.CULL_FACE,
					blend_func = (mgl.SRC_ALPHA, mgl.ONE),
					program = scene.ctx.program(
						vertex_shader=vert,
						fragment_shader=open(resourcedir+'/shaders/scheme-ghost.frag').read(),
						)),
				}
			return shaders, shader_ident
			
		def compute_spaces(self, view):
			''' Compute the new spaces for this frame.
				This is meant to be overriden when new spaces are required 
			'''
			view.uniforms['world'] = self.world
			for i,gen in enumerate(self.spacegens):
				self.spaces[i] = gen(view)
			if self.components:
				invview = affineInverse(view.uniforms['view'])
				for space,disp in self.components:
					disp.world = invview * self.spaces[space]
			self.b_spaces.write(self.spaces)
		
		def render(self, view):
			''' Render each va in self.vas '''
			ctx = view.scene.ctx
			for name in self.vas:
				if name not in self.shaders:
					continue
				shader = self.shaders[name]
				va = self.vas[name]
				ctx.enable_only(shader.enable)
				ctx.blend_func = shader.blend_func
				self.b_spaces.bind_to_storage_buffer(0)
				shader.program['b_spaces'].binding = 0
				shader.program['proj'].write(view.uniforms['proj'])
				shader.program['highlight'].write( fvec4(fvec3(settings.display['select_color_line']), 0.5) if self.selected else fvec4(0) )
				va.render()
			ctx.enable_only(mgl.BLEND | mgl.DEPTH_TEST)
		
		def identify(self, view):
			''' Render all the triangles and lines for identification '''
			self.b_spaces.bind_to_storage_buffer(0)
			self.shader_ident['b_spaces'].binding = 0
			self.shader_ident['startident'] = view.identstep(self.nidents)
			self.shader_ident['proj'].write(view.uniforms['proj'])
			
			if self.vai_lines:		self.vai_lines.render(mgl.LINES)
			if self.vai_triangles:	self.vai_triangles.render(mgl.TRIANGLES)
		
		def stack(self, scene):
			yield ((), 'screen', -1, self.compute_spaces)
			yield ((), 'screen', 2, self.render) 
			yield ((), 'ident', 2, self.identify)
			for space,disp in self.components:
				yield from disp.stack(scene)
				
		@writeproperty
		def selected(self, value):
			for space, disp in self.components:
				disp.selected = value
				
class SchemeInstance:
	def __init__(self, name, scheme, *kwargs):
		self.name = name
		self.scheme = scheme
		vars(self).update(kwargs)
	
	class display(Scheme.display):
		def __init__(self, scene, instance):
			self.instance = instance
			disp = scene.resource(id(instance.scheme), lambda: inst.scheme.display(scene, instance.scheme))
			vars(self).update(vars(disp))
			self.spaces = deepcopy(disp.spaces)
			
		def compute_spaces(self, view):
			''' Computes the new spaces for this frame.
				This is meant to be overriden when new spaces are required 
			'''
			view.uniforms['world'] = self.world
			for i,gen in enumerate(self.spacegens):
				self.spaces[i] = gen(view, self.instance)
			if self.components:
				invview = affineInverse(view.uniforms['view'])
				for space,disp in self.components:
					disp.world = invview * self.spaces[space]

# create standard spaces

def view(view):
	proj = view.uniforms['proj']
	return fmat4(1/proj[0][0],  0,0,0,
				0, 1/proj[1][1], 0,0,
				0,0,1,0,
				0,0,-1,1)

def screen(view):
	return fmat4(view.target.width/2,0,0,0,
				0,view.target.height/2,0,0,
				0,0,1,0,
				0,0,-1,1)

def ubiquity(view):
	#return translate(fvec3(0,0,-1)) * fmat4(fmat3(view.uniforms['view']) * fmat3(view.uniforms['world']))
	proj = view.uniforms['proj']
	f = - (proj[3][2]-proj[3][3]) / (proj[2][2] - proj[2][3])
	n = - (proj[3][2]+proj[3][3]) / (proj[2][2] + proj[2][3])
	d = 3*n
	e = proj * fvec4(1,1,d,1)
	e /= e[3]
	return translate(fvec3(0,0,d)) * fmat4(
				fmat3(1/e[1]) 
				* fmat3(view.uniforms['view']) 
				* fmat3(view.uniforms['world']))

def world(view):
	return view.uniforms['view'] * view.uniforms['world']

@dataclass
class halo_world:
	position: fvec3
	def __call__(self, view):
		center = view.uniforms['view'] * (view.uniforms['world'] * fvec4(position,1))
		m = fmat4(1)
		m[3] = center
		return m
	
@dataclass
class halo_view:
	position: fvec3 
	def __call__(self, view):
		center = view.uniforms['view'] * (view.uniforms['world'] * fvec4(self.position,1))
		proj = view.uniforms['proj']
		e = proj * fvec4(1,1,center.z,1)
		e /= e[3]
		m = fmat4(1/e[1])
		m[3] = center
		return m

@dataclass
class halo_screen:
	position: fvec3
	def __call__(self, view):
		center = view.uniforms['view'] * (view.uniforms['world'] * fvec4(self.position,1))
		e = view.uniforms['proj'] * fvec4(1,1,center.z,1)
		e /= e[3]
		m = fmat4(2/(e[1]*view.target.height))
		m[3] = center
		return m

@dataclass
class scale_screen:
	center: fvec3
	def __call__(self, view):
		m = view.uniforms['view'] * view.uniforms['world']
		e = view.uniforms['proj'] * fvec4(1,1,(m*self.center).z,1)
		e /= e[3]
		return m * translate(self.center) * scale(fvec3(2 / (e[1]*view.target.height)))

@dataclass
class scale_view:
	center: fvec3
	def __call__(self, view):
		m = view.uniforms['view'] * view.uniforms['world']
		e = view.uniforms['proj'] * fvec4(1,1,(m*self.center).z,1)
		e /= e[3]
		return m * translate(self.center) * scale(fvec3(1 / e[1]))

@dataclass
class halo_screen_flipping:
	position: fvec3
	offset: fvec3
	
	def __call__(self, view):
		numprec = 1e-7
		center = view.uniforms['view'] * (view.uniforms['world'] * fvec4(self.position,1))
		e = view.uniforms['proj'] * fvec4(1,1,center.z,1)
		e /= e[3]
		screen_center = view.uniforms['proj'] * center
		screen_center /= screen_center[3]
		screen_tip = view.uniforms['proj'] * (view.uniforms['view'] * (view.uniforms['world'] * fvec4(self.offset, 0)))
		screen_tip /= -abs(screen_tip[3]) - numprec
		if screen_tip[0] <= screen_center[0] + numprec:
			flip = fvec4(1,1,1,1)
		else:
			flip = fvec4(-1,1,1,1)
		m = fmat4(*(flip * 2/(e[1]*view.target.height)))
		m[3] = center
		return m

def mesh_placement(mesh) -> '(pos,normal)':
	''' Return an axis for placement of a note on the given object '''
	if isinstance(mesh, Mesh):
		# group center normal
		center = mesh.barycenter()
		f = min(mesh.faces,
				key=lambda f:  distance2(center, sum(mesh.facepoints(f))/3)
				)
		normal = mesh.facenormal(f)
		pos = sum(mesh.facepoints(f)) / 3
	
	elif isinstance(mesh, Web):
		center = mesh.barycenter()
		e = min(mesh.edges,
				key=lambda e:  distance2(center, (mesh.points[e[0]]+mesh.points[e[1]])/2)
				)
		normal = dirbase(normalize(mesh.points[e[0]]-mesh.points[e[1]]))[0]
		pos = mix(mesh.points[e[0]], mesh.points[e[1]], 0.5)
		
	elif isinstance(mesh, Wire):
		return mesh_placement(web(mesh))
		
	elif isaxis(mesh):
		pos, normal = mesh
	
	elif isinstance(mesh, vec3):
		normal = Z
		pos = mesh
		
	elif hasattr(mesh, 'mesh'):
		return mesh_placement(mesh.mesh())
	
	else:
		raise TypeError('unable to place note on a {}'.format(type(mesh)))
	return pos, normal


def note_floating(position, text, align=(0,0)):
	''' place a floating note at given position '''
	return Displayable(TextDisplay, position, text, align=align, color=settings.colors['annotation'], size=settings.display['view_font_size'])


def note_leading(placement, offset=None, text='here'):
	''' Place a leading note at given position
	
		`placement` can be any of `Mesh, Web, Wire, axis, vec3`
	'''
	origin, normal = mesh_placement(placement)
	if not offset:
		offset = 0.2 * length(boundingbox(placement).width) * normal
	elif isinstance(offset, (float,int)):
		offset = offset*normal
	elif not isinstance(offset, vec3):
		raise TypeError('offset must be scalar or vector')
	
	color = settings.colors['annotation']
	sch = Scheme(color=fvec4(color,0.7))
	sch.add([origin, origin+offset], shader='line', space=world)
	x,y,z = dirbase(normalize(vec3(offset)))
	sch.add(
		gt.revolution( 
			web([vec3(0), 16*z+4*x]), 
			Axis(vec3(0), z),
			resolution=('div',8),
			), 
		shader='fill',
		space=scale_screen(fvec3(origin)),
		)
	font = settings.display['view_font_size']
	sch.set(space=halo_screen_flipping(fvec3(origin+offset), fvec3(offset)))
	sch.add([vec3(0), vec3(font*2, 0, 0)], shader='line')
	sch.add(Displayable(TextDisplay, vec3(textsize(text)[0]/2*0.78*font + 2*font, 0, 0), text, align=('center', 0.5), color=fvec4(color,1), size=font))
	
	return sch

def note_distance(a, b, offset=0, project=None, d=None, tol=None, text=None, side=False):
	''' Place a distance quotation between 2 points, 
		the distance can be evaluated along vector `project` if specified 
	'''
	# get text to display
	if not project:					project = normalize(b-a)
	elif dot(project, b-a) < 0:		project = -project
	if not d:	d = abs(dot(b-a, project))
	if not text:
		if isinstance(tol,str): text = '{d:.4g}  {tol}'
		elif tol:               text = '{d:.4g}  ± {tol}'
		else:                   text = '{d:.4g}'
	text = text.format(d=d, tol=tol)
	color = settings.colors['annotation']
	# convert input vectors
	x = dirbase(project, b-a)[0]
	z = project if side else -project
	if not isinstance(offset, vec3):
		offset = offset * x
	shift = 0.5 * mathutils.project(b-a, x)	if length2(x) else 0
	ao = a + offset + shift
	bo = b + offset - shift
	# create scheme
	sch = Scheme()
	sch.set(shader='line', layer=1e-4, color=fvec4(color,0.3))
	sch.add([a, ao])
	sch.add([b, bo])
	sch.set(layer=-1e-4, color=fvec4(color,0.7))
	sch.add([ao, bo])
	sch.add(Displayable(TextDisplay,
				mix(ao,bo,0.5), 
				text, 
				align=('center', 0.5), 
				size=settings.display['view_font_size'], 
				color=fvec4(color,1)))
	sch.set(shader='fill')
	sch.add(gt.revolution(
				web([vec3(0), 3*x-12*z]), 
				Axis(vec3(0),project), 
				resolution=('div',8)), 
			space=scale_screen(fvec3(ao)))
	sch.add(gt.revolution(
				web([vec3(0), 3*x+12*z]), 
				Axis(vec3(0),project), 
				resolution=('div',8)), 
			space=scale_screen(fvec3(bo)))
	return sch
	
def note_distance_planes(s0, s1, offset=None, d=None, tol=None, text=None):
	''' Place a distance quotation between 2 meshes 
		
		`s0` and `s1` can be any of `Mesh, Web, Wire`
	'''
	p0, n0 = mesh_placement(s0)
	p1, n1 = mesh_placement(s1)
	if length2(cross(n0,n1)) > 1e-10:
		raise ValueError('surfaces are not parallel')
	if not offset:
		box = boundingbox(s0,s1)
		offset = 0.2 * length(box.width) * normalize(noproject(p0+p1 - 2*box.center, n0))
	return note_distance(p0, p1, offset, n0, d, tol, text)
	
def note_distance_set(s0, s1, offset=0, d=None, tol=None, text=None):
	''' Place a distance quotation between 2 objects. This is the distance between the closest elements of both sets 
	
		`s0` and `s1` can be any of `Mesh, Web, Wire, vec3`
	'''
	dist, p0, p1 = mesh_distance(s0, s1)
	if not d:	d = dist
	# take the point and the higher dimension primitive
	if not isinstance(p0,int):	s0,p0, s1,p1 = s1,p1, s0,p0
	p0 = s0.points[p0]
	pts = s1.points
	if isinstance(p1, tuple):
		if len(p1) == 2:	p1 = pts[p1[0]] + project(p0-pts[p1[0]], pts[p1[1]]-pts[p1[0]])
		elif len(p1) == 3:	p1 = p0 + project(pts[p1[0]]-p0, cross(pts[p1[1]]-pts[p1[0]], pts[p1[2]]-pts[p1[0]]))
	else:
		p1 = s1.points[p1]
	
	return note_distance(p0, p1, offset, None, d, tol, text)
		

def note_angle(a0, a1, offset=0, d=None, tol=None, text=None, unit='deg', side=False):
	''' Place an angle quotation between 2 axis '''
	o0, d0 = a0
	o1, d1 = a1
	z = normalize(cross(d0,d1))
	if not isfinite(z):	
		raise ValueError('axis are parallel')
	x0 = cross(d0,z) * (1 if side else -1)
	x1 = cross(d1,z) * (1 if side else -1)
	shift = project(o1-o0, z) * 0.5
	# add it but in a new copy of the vectors
	o0 = o0 + shift
	o1 = o1 + shift
	# get text to display
	if not d:	d = anglebt(d0, d1)
	if unit == 'deg':
		d *= 180/pi
		unit = '°'
	if not text:
		if isinstance(tol,str): text = '{d:.4g}{unit}  {tol}'
		elif tol:               text = '{d:.4g}{unit}  ± {tol}'
		else:                   text = '{d:.4g}{unit}'
	text = text.format(d=d, tol=tol, unit=unit)
	color = settings.colors['annotation']
	# arc center
	if o1 == o0:	center = o0
	else:			center = o0 + unproject(project(o1-o0, x1), d0)
	radius = mix(distance(o0,center), distance(o1,center), 0.5) + offset
	if not radius:	radius = 1
	# arc extremities
	p0 = center+radius*d0
	p1 = center+radius*d1
	sch = Scheme()
	sch.set(shader='line', layer=1e-4, color=fvec4(color,0.3))
	sch.add([p0, o0])
	sch.add([p1, o1])
	arc = ArcCentered((center,z), p0, p1, ('rad',0.05)).mesh()
	sch.add(arc, color=fvec4(color,0.7))
	sch.set(layer=-1e-4)
	sch.add(Displayable(TextDisplay,
				arc[len(arc)//2], 
				text, 
				align=('center', 0.5), 
				size=settings.display['view_font_size'], 
				color=fvec4(color,1)))
	sch.set(shader='fill')
	sch.add(gt.revolution(
				web([vec3(0), 3*d0+12*x0]), 
				Axis(vec3(0),x0), 
				resolution=('div',8)), 
			space=scale_screen(fvec3(p0)))
	sch.add(gt.revolution(
				web([vec3(0), 3*d1-12*x1]), 
				Axis(vec3(0),x1), 
				resolution=('div',8)), 
			space=scale_screen(fvec3(p1)))
	return sch
	
def note_radius(mesh, offset=None, d=None, tol=None, text=None, propagate=2):
	''' Place a curvature radius quotation. This will be the minimal curvature radius observed in the mesh 
		As a mesh is generally speaking an approximation of the desired shape, the radius may be approximative as well
	'''
	if isinstance(mesh, Mesh):
		normals = mesh.vertexnormals()
		radius, place = mesh_curvature_radius(mesh, normals=normals, propagate=propagate)
		normal = normals[place]
	elif isinstance(mesh, Web):
		conn = connpp(mesh.edges)
		radius, place = mesh_curvature_radius(mesh, conn=conn, propagate=propagate)
		normal = - normalize(sum(normalize(mesh.points[p]-mesh.points[place])  for p in conn[place]))
	elif isinstance(mesh, wire):
		normals = mesh.vertexnormals()
		radius, place = mesh_curvature_radius(mesh, normals=normals, propagate=propagate)
		normal = normals[place]
	else:
		raise TypeError('input mesh must be Mesh, Web or Wire')
	if not offset:
		offset = 0.2 * length(boundingbox(mesh).width)
	if not isinstance(offset, vec3):
		offset = normal * offset
	if abs(dot(offset, normal))**2 < length2(place)*1e-4:
		offset += 0.2 * length(boundingbox(mesh).width) * normal
	
	if isinstance(place, tuple):
		place = sum(mesh.points[i]  for i in place) / len(place)
	else:
		place = mesh.points[place]
	
	arrowplace = place + noproject(offset, normal)
	note = note_leading((arrowplace,normal), offset=project(offset, normal), text=text or 'R {:.4g}'.format(radius))
	color = settings.colors['annotation']
	note.set(shader='line', layer=1e-4, color=fvec4(color,0.3), space=world)
	note.add([place, arrowplace])
	return note
	
	
def mesh_curvature_radius(mesh, conn=None, normals=None, propagate=2) -> '(distance, point)':
	''' Find the minimum curvature radius of a mesh.
	
		Parameters:
		
			mesh:			the surface/line to search
			conn:			a point-to-point connectivity (computed if not provided)
			normals:		the vertex normals (computed if not provided)
			propagate(int):	the maximum propagation rank for points to pick for the regression
	
		Return:	`(distance: float, point: int)` where primitives varies according to the input mesh dimension
	'''
		
	def propagate_pp(conn, start, maxrank):
		front = [(0,s) for s in start]
		seen = set()
		while front:
			rank, p = front.pop()
			if p in seen:	continue
			seen.add(p)
			yield p
			if rank < maxrank:
				for n in conn[p]:
					if n not in seen:	
						front.append((rank+1, n))
	
	if isinstance(mesh, Mesh):		
		if not conn:	conn = connpp(mesh.faces)
		if not normals:	normals = mesh.vertexnormals()
		it = ( (p, list(propagate_pp(conn, [p], propagate)))   for p in connpp )
	elif isinstance(mesh, Web):	
		if not conn:	conn = connpp(mesh.edges)
		if not normals:	
			normals = [vec3(0)  for p in mesh.points]
			pts = mesh.points
			for e in mesh.edges:
				d = pts[e[0]] - pts[e[1]]
				normals[e[0]] += d
				normals[e[1]] -= d
			for i,n in enumerate(normals):
				normals[i] = normalize(n)
		it = ( (p, list(propagate_pp(conn, [p], propagate)))   for p in connpp )
	elif isinstance(mesh, Wire):
		if not normals:	normals = mesh.vertexnormals()
		it = ( (mesh.indices[i], mesh.indices[i-propagate:i+propagate])   for i in range(len(mesh.indices)) )
	else:
		raise TypeError('bad input type')
	
	def analyse():
		pts = mesh.points		
		for p, neigh in it:
			# decide a local coordinate system
			u,v,w = dirbase(normals[p])
			# get neighboors contributions
			b = np.empty(len(neigh))
			a = np.empty((len(neigh), 14))
			for i,n in enumerate(neigh):
				e = pts[n] - pts[p]
				b[i] = dot(e,w)
				eu = dot(e,u)
				ev = dot(e,v)
				# these are the monoms to compose to build a polynom approximating the surface until 4th-order derivatives
				a[i] = (	eu**2, ev**2, eu*ev,
							eu, ev,
							eu**3, eu**2*ev, eu*ev**2, ev**3,
							eu**4, eu**3*ev, eu**2*ev**2, eu*ev**3, ev**4,
							)

			# least squares resulution, the complexity is roughly the same as inverting a mat3
			(au, av, auv, *_), residuals, *_ = np.linalg.lstsq(a, b)
			# diagonalize the curve tensor to get the principal curvatures
			diag, transfer = np.linalg.eigh(mat2(2*au, auv, 
													auv, 2*av))
			yield 1/np.max(np.abs(diag)), p
	
	return min(analyse(), key=itemgetter(0), default=None)
	
	
def mesh_curvature_radius(mesh, conn=None, normals=None, propagate=2) -> '(distance, point)':
	''' Find the minimum curvature radius of a mesh.
	
		Parameters:
		
			mesh:			the surface/line to search
			conn:			a point-to-point connectivity (computed if not provided)
			normals:		the vertex normals (computed if not provided)
			propagate(int):	the maximum propagation rank for points to pick for the regression
	
		Return:	`(distance: float, point: int)` where primitives varies according to the input mesh dimension
	'''
		
	curvatures = mesh_curvatures(mesh, conn, normals, propagate)
	
	place = min(range(len(mesh.points)),
				key=lambda p: 1/np.max(np.abs(curvatures[p][0]))  if curvatures[p] else inf, 
				default=None)
	return 1/np.max(np.abs(curvatures[place][0])), place
	
def mesh_curvatures(mesh, conn=None, normals=None, propagate=2):
	''' Compute the curvature around a point in a mesh/web/wire
	
		Parameters:
		
			mesh:			the surface/line to search
			conn:			a point-to-point connectivity (computed if not provided)
			normals:		the vertex normals (computed if not provided)
			propagate(int):	the maximum propagation rank for points to pick for the regression
	
		Return:	
			
			`[(tuple, mat3)]`
			
			where the `tuple` contains the curvature in each of the column directions in the `mat3`. The `mat3` has the principal directions of curvature
	'''
	pts = mesh.points
		
	# propagate though the mesh and return seen points
	def propagate_pp(conn, start, maxrank):
		front = deque((0,s) for s in start)
		seen = set()
		while front:
			rank, p = front.popleft()
			if p in seen:	continue
			seen.add(p)
			if rank < maxrank:
				for n in conn[p]:
					if n not in seen:	
						front.append((rank+1, n))
		return seen
	
	if isinstance(mesh, Mesh):		
		if not conn:	conn = connpp(mesh.faces)
		if not normals:	normals = mesh.vertexnormals()
		it = ( (p, propagate_pp(conn, [p], propagate))   for p in conn )
	elif isinstance(mesh, Web):	
		if not conn:	conn = connpp(mesh.edges)
		if not normals:	
			normals = [vec3(0)  for p in mesh.points]
			for e in mesh.edges:
				d = pts[e[0]] - pts[e[1]]
				normals[e[0]] += d
				normals[e[1]] -= d
			for i,n in enumerate(normals):
				normals[i] = normalize(n)
		it = ( (p, propagate_pp(conn, [p], propagate))   for p in conn )
	elif isinstance(mesh, Wire):
		if not normals:	normals = mesh.vertexnormals()
		it = ( (mesh.indices[i], mesh.indices[i-propagate:i+propagate])   for i in range(len(mesh.indices)) )
	else:
		raise TypeError('bad input shape type')
	
	curvatures = [None]*len(pts)
	
	for p, neigh in it:
		# decide a local coordinate system
		u,v,w = dirbase(normals[p])
		# get neighboors contributions
		b = np.empty(len(neigh))
		a = np.empty((len(neigh), 14))
		for i,n in enumerate(neigh):
			e = pts[n] - pts[p]
			b[i] = dot(e,w)
			eu = dot(e,u)
			ev = dot(e,v)
			# these are the monoms to compose to build a polynom approximating the surface until 4th-order derivatives
			a[i] = (	eu**2, ev**2, eu*ev,  # what we want to get
						# the others terms are only present to catch the surface regularities and not disturb the curvature terms
						eu, ev,
						eu**3, eu**2*ev, eu*ev**2, ev**3,
						eu**4, eu**3*ev, eu**2*ev**2, eu*ev**3, ev**4,
						)

		# least squares resulution, the complexity is roughly the same as inverting a mat3
		(au, av, auv, *_), residuals, *_ = np.linalg.lstsq(a, b, rcond=None)
		# diagonalize the curve tensor to get the principal curvatures
		diag, transfer = np.linalg.eigh(mat2(2*au, auv, 
												auv, 2*av))
		curvatures[p] = diag, mat3(u,v,w) * mat3(mat2(transfer))
		
	return curvatures


def note_bounds(obj):
	''' Create dimension annotations on the boundingbox of an object '''
	box = boundingbox(obj)
	size = 0.05 * length(box.width)
	return [
		note_distance(box.min, vec3(box.max.x, box.min.y, box.min.z), offset=size*vec3(0,-1,-1)),
		note_distance(box.min, vec3(box.min.x, box.max.y, box.min.z), offset=size*vec3(-1,0,-1)),
		note_distance(box.min, vec3(box.min.x, box.min.y, box.max.z), offset=size*vec3(-1,-1,0)),
		box,
		]


	
def _mesh_direction(mesh):
	if isinstance(mesh, Mesh):	return mesh.facenormal(0)
	elif isinstance(mesh, Web):	return mesh.edgedirection(0)
	elif isinstance(mesh, Wire):	return mesh[1]-mesh[0]
	elif isaxis(mesh):	return mesh[1]
	else:
		raise TypeError('only Mesh and Web are supported')
	
def note_angle_planes(s0, s1, offset=0, d=None, tol=None, text=None, unit='deg'):
	''' Place an angle quotation between 2 meshes considered to be plane (surface) or straight (curve) 
	
		`s0` and `s1` can be any of `Mesh, Web, Wire, Axis`
	'''
	d0, d1 = _mesh_direction(s0), _mesh_direction(s1)
	z = normalize(cross(d0, d1))
	if not isfinite(z):	
		raise ValueError('planes are parallel')
	if isinstance(s0, Mesh) or isaxis(s0):	d0 = cross(d0,z)
	if isinstance(s1, Mesh) or isaxis(s1):	d1 = cross(z,d1)
	return note_angle(
				(mesh_placement(s0)[0], d0), 
				(mesh_placement(s1)[0], d1), 
				offset, d, tol, text, unit)
				
def note_angle_edge(part, edge, offset=0, d=None, tol=None, text=None, unit='deg'):
	''' Place an angle quotation around a mesh edge '''
	f0 = None
	f1 = None
	for face in part.faces:
		for i in range(3):
			if face[i-1] == edge[0] and face[i] == edge[1]:	
				f0 = face
			elif face[i-1] == edge[1] and face[i] == edge[0]:
				f1 = face
	if not f0 or not f1:
		raise ValueError("edge {} doesn't exist or is not between 2 faces".format(f0))
	d0 = part.facenormal(f0)
	d1 = part.facenormal(f1)
	z = normalize(cross(d0,d1))
	o = mix(part.points[edge[0]], part.points[edge[1]], 0.5)
	if not offset:	offset = 0.2 * length(boundingbox(part).width)
	return note_angle(
			(o, cross(z,d0)),
			(o, cross(d1,z)),
			offset, d, tol, text, unit)
	
def note_absciss(axis, pts):
	indev
	
def note_diameter(mesh, direction=None, offset=None, d=None, tol=None, text=None):
	assert direction is not None
	for f in mesh.faces:
		for p in f:
			if dot(mesh.points[p], direction):
				indev

def note_surface(placement, offset=None, roughness=None, method=None):
	indev
	
def note_label(placement, offset=None, text='!', style='rect'):
	''' Place a text label upon an object 
	
		`placement` can be any of `Mesh, Web, Wire, axis, vec3`
	'''
	p, normal = mesh_placement(placement)
	if not offset:
		offset = 0.2 * length(boundingbox(placement).width) * normal
	elif isinstance(offset, (float,int)):
		offset = offset*normal
	elif not isinstance(offset, vec3):
		raise TypeError('offset must be scalar or vector')
	
	color = settings.colors['annotation']
	x,_,z = dirbase(normalize(offset))
	sch = Scheme()
	sch.set(color=fvec4(color,0.7))
	sch.add(gt.revolution(
				web([6*x, 10*z]),
				Axis(vec3(0),z),
				resolution=('div',8),
				),
			space=scale_screen(fvec3(p)), shader='fill')
	sch.add([p, p+offset], space=world, shader='line', layer=2e-4)
	font = int(settings.display['view_font_size'] * 1.2)
	r = font*0.8
	print(font, r)
	if style == 'circle':	outline = web(Circle((vec3(0),vec3(0,0,1)), r))
	elif style == 'rect':	outline = [vec3(r,r,0), vec3(-r,r,0), vec3(-r,-r,0), vec3(r,-r,0), vec3(r,r,0)]
	else:
		raise ValueError("style must be 'rect' or 'circle'")
	sch.set(space=halo_screen(fvec3(p+offset)))
	sch.add(outline, shader='line', layer=0)
	sch.add(gt.flatsurface(wire(outline)), color=fvec4(settings.display['background_color'],0), shader='fill', layer=1e-3)
	sch.add(Displayable(TextDisplay,
				p+offset, 
				text, 
				align=('center','center'), 
				size=font, 
				color=color), 
			space=world)
	return sch
	

def note_iso(p, offset, type, text, refs=(), label=None):
	indev



#from .scheme import SchemeInstance

#class BaseDisplay(Display):
	#scheme = None
	
	#def __init__(self, scene, matrix, color=None):
		#if not color:	color = settings.colors['annotation']
		
		#if not self.scheme:
			#type(self).scheme = sch = Scheme(layer=1e-4)
			#space = lambda view, instance: world(view) * instance.matrix
			
			#directions = [vec3(1,0,0), vec3(0,1,0), vec3(0,0,1)]
			#for i in range(len(directions)):
				#o = vec3(0)
				#x,z = directions[i-1], directions[i]
				
				#def tip(view, instance, z=z): 
					#s = space(view, instance)
					#return scale_screen(s*z) * fmat4(fmat3(s))
				
				#sch.add([o, z], 
					#space=tip, 
					#color=fvec4(color,1), 
					#shader='line',
					#)
				#sch.add(revolution(2*pi, (o,z), web([o, -20*z+9*x])), 
					#space=tip, 
					#color=fvec4(color,0.8), 
					#shader='fill',
					#)
		
		#if isinstance(matrix, (dmat3,fmat3,dquat,fquat)):	matrix = fmat4(fmat3(matrix))
		#elif isinstance(matrix, fmat4):						matrix = fmat4(matrix)
		#self.disp = scene.display(SchemeInstance(self.scheme, matrix=matrix))
	
	#def stack(self, scene):
		#yield from self.disp.stack(scene)
		
def note_base(matrix, color=None, labels=('X', 'Y', 'Z'), name=None, size=0.4):
	if not color:	color = settings.colors['annotation']
	if not name:	name = ''
	if not labels:	labels = [None]*3
	
	if isinstance(matrix, (dmat3,fmat3)):
		center = None
		directions = mat3(matrix)
	elif isinstance(matrix, (dmat4,fmat4)):
		if length2(transpose(matrix)[3] - vec4(0,0,0,1)) > NUMPREC:
			print(matrix, length2(transpose(matrix)[3] - vec4(0,0,0,1)))
			raise TypeError('mat4 must be affine in order to be displayed')
		center = vec3(matrix[3])
		directions = mat3(mat4(matrix))
	else:
		raise TypeError('only matrix with dimension 3 or 4 can be displayed')
	
	sch = Scheme(layer=1e-4)
	for axislabel, direction in zip(labels, directions):
		o = vec3(0)
		x,y,z = dirbase(normalize(direction))
		
		if center:
			sch.add([center], space=world)
		
		sch.add([o, size*direction], 
			space=scale_view(fvec3(center)) if center else ubiquity, 
			color=fvec4(color,1), 
			shader='line',
			)
		sch.add(gt.revolution(web([o, -14*z+3.6*x]), Axis(o,z)), 
			space=scale_screen_scale_view(fvec3(center), fvec3(size*direction)) if center else scale_screen_ubiquity(fvec3(size*direction)), 
			color=fvec4(color,0.8), 
			shader='fill',
			)
		if axislabel or name:
			sch.add(Displayable(TextDisplay, 5*z, axislabel+' '+name, align=(0.5,0.5), color=color))
	
	return sch
	
def note_rotation(quaternion, color=None, size=0.4):
	if not color:	color = settings.colors['annotation']
	
	direction = vec3(glm.axis(quaternion))
	angle = glm.angle(quaternion)
	o = vec3(0)
	x,y,z = dirbase(direction)
	tend = -sin(angle)*x + cos(angle)*y
	rend = cos(angle)*x + sin(angle)*y
	
	sch = Scheme(layer=1e-4)
	sch.set(
		space=ubiquity,
		color=fvec4(color,0.6),
		shader='line',
		)
	sch.add([-size*direction, -0.2*size*direction])
	sch.add([-0.1*size*direction, 0.1*size*direction])
	sch.add([0.2*size*direction, size*direction])
	sch.add(gt.revolution(web([o, -14*z+3.6*x]), Axis(o,z)), 
		space=scale_screen_ubiquity(fvec3(size*direction)), 
		color=fvec4(color,0.6), 
		shader='fill',
		)
	
	sch.set(
		space=ubiquity,
		shader='line',
		)
	sch.add([size*cos(t)*x + size*sin(t)*y  for t in linrange(0, angle, step=0.1)], color=fvec4(color,1))
	sch.add([size*cos(t)*x + size*sin(t)*y  for t in linrange(angle, 2*pi, step=0.1)], color=fvec4(color,0.2))
	sch.add(gt.revolution(web([o, -14*tend-3.6*rend]), Axis(o, tend), resolution=('div', 8)), 
		space=scale_screen_ubiquity(fvec3(size*rend)), 
		color=fvec4(color,0.8), 
		shader='fill',
		)
		
	return sch
	
'''
transforms = [
	mat3(),
	angleAxis(pi/2, normalize(Z+X)),
	translate(2*Y) * scale(vec3(0.2, 0.3, 1)),
	translate(2*X),
	rotate(1, vec3(1,2,3)),
	translate(-2*X) * mat4(scaledir(vec3(1,2,-3)*0.1)),
	]
'''
	
	
def scale_screen_scale_view(center, offset):
	def mat(view):
		m = view.uniforms['view'] * view.uniforms['world']
		e = view.uniforms['proj'] * fvec4(1,1,(m*center).z,1)
		e /= e[3]
		return m * translate(center) * scale(fvec3(1/e[1])) * translate(offset) * scale(fvec3(2/view.target.height))
	return mat
	
def scale_screen_ubiquity(offset):
	def mat(view):
		proj = view.uniforms['proj']
		f = - (proj[3][2]-proj[3][3]) / (proj[2][2] - proj[2][3])
		n = - (proj[3][2]+proj[3][3]) / (proj[2][2] + proj[2][3])
		d = 3*n
		e = proj * fvec4(1,1,d,1)
		e /= e[3]
		d2 = 2/view.target.height
		return translate(fvec3(0,0,d)) * fmat4(
					fmat3(1/e[1]) 
					* fmat3(view.uniforms['view']) 
					* fmat3(view.uniforms['world'])) * translate(offset) * scale(fvec3(d2))
	return mat
	
def base_display(scene, mat):
	return scene.display(note_base(mat))
	#try:	return scene.display(note_base(mat))
	#except TypeError as err:	return Display()

def quat_display(scene, quat):
	try:	return scene.display(note_rotation(quat))
	except TypeError:	return Display()
		
overrides.update({
	dmat4: base_display,
	fmat4: base_display,
	dmat3: base_display,
	fmat3: base_display,
	dquat: quat_display,
	fquat: quat_display,
	})
